import numpy as np

import pandas as pd

import pytest
from pandas.api.types import is_numeric_dtype

from scipy import sparse

import sklearn.datasets
import sklearn.model_selection
from sklearn.utils.multiclass import type_of_target

from autosklearn.data.target_validator import TargetValidator


# Fixtures to be used in this class. By default all elements have 100 datapoints
@pytest.fixture
def input_data_targettest(request):
    if request.param == 'series_binary':
        return pd.Series([1, -1, -1, 1])
    elif request.param == 'series_multiclass':
        return pd.Series([1, 0, 2])
    elif request.param == 'series_multilabel':
        return pd.Series([[1, 0], [0, 1]])
    elif request.param == 'series_continuous':
        return pd.Series([0.1, 0.6, 0.7])
    elif request.param == 'series_continuous-multioutput':
        return pd.Series([[1.5, 2.0], [3.0, 1.6]])
    elif request.param == 'pandas_binary':
        return pd.DataFrame([1, -1, -1, 1])
    elif request.param == 'pandas_multiclass':
        return pd.DataFrame([1, 0, 2])
    elif request.param == 'pandas_multilabel':
        return pd.DataFrame([[1, 0], [0, 1]])
    elif request.param == 'pandas_continuous':
        return pd.DataFrame([0.1, 0.6, 0.7])
    elif request.param == 'pandas_continuous-multioutput':
        return pd.DataFrame([[1.5, 2.0], [3.0, 1.6]])
    elif request.param == 'numpy_binary':
        return np.array([1, -1, -1, 1])
    elif request.param == 'numpy_multiclass':
        return np.array([1, 0, 2])
    elif request.param == 'numpy_multilabel':
        return np.array([[1, 0], [0, 1]])
    elif request.param == 'numpy_continuous':
        return np.array([0.1, 0.6, 0.7])
    elif request.param == 'numpy_continuous-multioutput':
        return np.array([[1.5, 2.0], [3.0, 1.6]])
    elif request.param == 'list_binary':
        return [1, -1, -1, 1]
    elif request.param == 'list_multiclass':
        return [1, 0, 2]
    elif request.param == 'list_multilabel':
        return [[0, 1], [1, 0]]
    elif request.param == 'list_continuous':
        return [0.1, 0.6, 0.7]
    elif request.param == 'list_continuous-multioutput':
        return [[1.5, 2.0], [3.0, 1.6]]
    elif 'openml' in request.param:
        _, openml_id = request.param.split('_')
        X, y = sklearn.datasets.fetch_openml(data_id=int(openml_id),
                                             return_X_y=True, as_frame=True)
        if len(y.shape) > 1 and y.shape[1] > 1 and np.any(y.eq('TRUE').any(1).to_numpy()):
            # This 'if' is only asserted for multi-label data
            # Force the downloaded data to be interpreted as multilabel
            y = y.dropna()
            y.replace('FALSE', 0, inplace=True)
            y.replace('TRUE', 1, inplace=True)
            y = y.astype(int)
        return y
    elif 'sparse' in request.param:
        # We expect the names to be of the type sparse_csc_nonan
        sparse_, type_, nan_ = request.param.split('_')
        if 'nonan' in nan_:
            data = np.ones(3)
        else:
            data = np.array([1, 2, np.nan])

        # Then the type of sparse
        if 'csc' in type_:
            return sparse.csc_matrix(data)
        elif 'csr' in type_:
            return sparse.csr_matrix(data)
        elif 'coo' in type_:
            return sparse.coo_matrix(data)
        elif 'bsr' in type_:
            return sparse.bsr_matrix(data)
        elif 'lil' in type_:
            return sparse.lil_matrix(data)
        elif 'dok' in type_:
            return sparse.dok_matrix(np.vstack((data, data, data)))
        elif 'dia' in type_:
            return sparse.dia_matrix(np.vstack((data, data, data)))
        else:
            ValueError("Unsupported indirect fixture {}".format(request.param))
    else:
        ValueError("Unsupported indirect fixture {}".format(request.param))


# Actual checks for the targets
@pytest.mark.parametrize(
    'input_data_targettest',
    (
        'series_binary',
        'series_multiclass',
        'series_continuous',
        'pandas_binary',
        'pandas_multiclass',
        'pandas_multilabel',
        'pandas_continuous',
        'pandas_continuous-multioutput',
        'numpy_binary',
        'numpy_multiclass',
        'numpy_multilabel',
        'numpy_continuous',
        'numpy_continuous-multioutput',
        'list_binary',
        'list_multiclass',
        'list_multilabel',
        'list_continuous',
        'list_continuous-multioutput',
        'sparse_bsr_nonan',
        'sparse_coo_nonan',
        'sparse_csc_nonan',
        'sparse_csr_nonan',
        'sparse_lil_nonan',
        'openml_204',
    ),
    indirect=True
)
def test_targetvalidator_supported_types_noclassification(input_data_targettest):
    validator = TargetValidator(is_classification=False)
    validator.fit(input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    if sparse.issparse(input_data_targettest):
        assert sparse.issparse(transformed_y)
    else:
        assert isinstance(transformed_y, np.ndarray)
    epected_shape = np.shape(input_data_targettest)
    if len(epected_shape) > 1 and epected_shape[1] == 1:
        # The target should have (N,) dimensionality instead of (N, 1)
        epected_shape = (epected_shape[0], )
    assert epected_shape == np.shape(transformed_y)
    assert np.issubdtype(transformed_y.dtype, np.number)
    assert validator._is_fitted

    # Because there is no classification, we do not expect a encoder
    assert validator.encoder is None

    if hasattr(input_data_targettest, "iloc"):
        np.testing.assert_array_equal(
            np.ravel(input_data_targettest.to_numpy()),
            np.ravel(transformed_y)
        )
    elif sparse.issparse(input_data_targettest):
        np.testing.assert_array_equal(
            np.ravel(input_data_targettest.todense()),
            np.ravel(transformed_y.todense())
        )
    else:
        np.testing.assert_array_equal(
            np.ravel(np.array(input_data_targettest)),
            np.ravel(transformed_y)
        )


@pytest.mark.parametrize(
    'input_data_targettest',
    (
        'series_binary',
        'series_multiclass',
        'pandas_binary',
        'pandas_multiclass',
        'numpy_binary',
        'numpy_multiclass',
        'list_binary',
        'list_multiclass',
        'sparse_bsr_nonan',
        'sparse_coo_nonan',
        'sparse_csc_nonan',
        'sparse_csr_nonan',
        'sparse_lil_nonan',
        'openml_2',
    ),
    indirect=True
)
def test_targetvalidator_supported_types_classification(input_data_targettest):
    validator = TargetValidator(is_classification=True)
    validator.fit(input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    if sparse.issparse(input_data_targettest):
        assert sparse.issparse(transformed_y)
    else:
        assert isinstance(transformed_y, np.ndarray)
    epected_shape = np.shape(input_data_targettest)
    if len(epected_shape) > 1 and epected_shape[1] == 1:
        # The target should have (N,) dimensionality instead of (N, 1)
        epected_shape = (epected_shape[0], )
    assert epected_shape == np.shape(transformed_y)
    assert np.issubdtype(transformed_y.dtype, np.number)
    assert validator._is_fitted

    # Because there is no classification, we do not expect a encoder
    if not sparse.issparse(input_data_targettest):
        assert validator.encoder is not None

        # The encoding should be per column
        if len(transformed_y.shape) == 1:
            assert np.min(transformed_y) == 0
            assert np.max(transformed_y) == len(np.unique(transformed_y)) - 1
        else:
            for col in range(transformed_y.shape[1]):
                assert np.min(transformed_y[:, col]) == 0
                assert np.max(transformed_y[:, col]) == len(np.unique(transformed_y[:, col])) - 1

        # Make sure we can perform inverse transform
        y_inverse = validator.inverse_transform(transformed_y)
        if hasattr(input_data_targettest, 'dtype'):
            # In case of numeric, we need to make sure dtype is preserved
            if is_numeric_dtype(input_data_targettest.dtype):
                assert y_inverse.dtype == input_data_targettest.dtype
            # Then make sure every value is properly inverse-transformed
            np.testing.assert_array_equal(np.array(y_inverse), np.array(input_data_targettest))
        elif hasattr(input_data_targettest, 'dtypes'):
            if is_numeric_dtype(input_data_targettest.dtypes[0]):
                assert y_inverse.dtype == input_data_targettest.dtypes[0]
            # Then make sure every value is properly inverse-transformed
            np.testing.assert_array_equal(np.array(y_inverse),
                                          # pandas is always (N, 1) but targets are ravel()
                                          input_data_targettest.to_numpy().reshape(-1))
    else:
        # Sparse is not encoded, mainly because the sparse data is expected
        # to be numpy of numerical type -- which currently does not require encoding
        np.testing.assert_array_equal(
            np.ravel(input_data_targettest.todense()),
            np.ravel(transformed_y.todense())
        )


@pytest.mark.parametrize(
    'input_data_targettest',
    (
        'series_binary',
        'pandas_binary',
        'numpy_binary',
        'list_binary',
        'openml_1066',
    ),
    indirect=True
)
def test_targetvalidator_binary(input_data_targettest):
    assert type_of_target(input_data_targettest) == 'binary'
    validator = TargetValidator(is_classification=True)
    # Test the X_test also!
    validator.fit(input_data_targettest, input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    assert type_of_target(transformed_y) == 'binary'


@pytest.mark.parametrize(
    'input_data_targettest',
    (
        'series_multiclass',
        'pandas_multiclass',
        'numpy_multiclass',
        'list_multiclass',
        'openml_54',
    ),
    indirect=True
)
def test_targetvalidator_multiclass(input_data_targettest):
    assert type_of_target(input_data_targettest) == 'multiclass'
    validator = TargetValidator(is_classification=True)
    # Test the X_test also!
    validator.fit(input_data_targettest, input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    assert type_of_target(transformed_y) == 'multiclass'


@pytest.mark.parametrize(
    'input_data_targettest',
    (
        'pandas_multilabel',
        'numpy_multilabel',
        'list_multilabel',
        'openml_40594',
    ),
    indirect=True
)
def test_targetvalidator_multilabel(input_data_targettest):
    assert type_of_target(input_data_targettest) == 'multilabel-indicator'
    validator = TargetValidator(is_classification=True)
    # Test the X_test also!
    validator.fit(input_data_targettest, input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    assert type_of_target(transformed_y) == 'multilabel-indicator'


@pytest.mark.parametrize(
    'input_data_targettest',
    (
        'series_continuous',
        'pandas_continuous',
        'numpy_continuous',
        'list_continuous',
        'openml_531',
    ),
    indirect=True
)
def test_targetvalidator_continuous(input_data_targettest):
    assert type_of_target(input_data_targettest) == 'continuous'
    validator = TargetValidator(is_classification=False)
    # Test the X_test also!
    validator.fit(input_data_targettest, input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    assert type_of_target(transformed_y) == 'continuous'


@pytest.mark.parametrize(
    'input_data_targettest',
    (
        'pandas_continuous-multioutput',
        'numpy_continuous-multioutput',
        'list_continuous-multioutput',
        'openml_41483',
    ),
    indirect=True
)
def test_targetvalidator_continuous_multioutput(input_data_targettest):
    assert type_of_target(input_data_targettest) == 'continuous-multioutput'
    validator = TargetValidator(is_classification=False)
    # Test the X_test also!
    validator.fit(input_data_targettest, input_data_targettest)
    transformed_y = validator.transform(input_data_targettest)
    assert type_of_target(transformed_y) == 'continuous-multioutput'


@pytest.mark.parametrize(
    'input_data_targettest',
    (
        'series_binary',
        'pandas_binary',
        'numpy_binary',
        'list_binary',
    ),
    indirect=True
)
def test_targetvalidator_fitontypeA_transformtypeB(input_data_targettest):
    """
    Check if we can fit in a given type (numpy) yet transform
    if the user changes the type (pandas then)

    This is problematic only in the case we create an encoder
    """
    validator = TargetValidator(is_classification=True)
    validator.fit(input_data_targettest)
    if isinstance(input_data_targettest, pd.DataFrame):
        complementary_type = input_data_targettest.to_numpy()
    elif isinstance(input_data_targettest, pd.Series):
        complementary_type = pd.DataFrame(input_data_targettest)
    elif isinstance(input_data_targettest, np.ndarray):
        complementary_type = pd.DataFrame(input_data_targettest)
    elif isinstance(input_data_targettest, list):
        complementary_type = pd.DataFrame(input_data_targettest)
    validator.transform(complementary_type)


@pytest.mark.parametrize(
    'input_data_targettest',
    (
        'series_multilabel',
        'series_continuous-multioutput',
    ),
    indirect=True
)
def test_type_of_target_unsupported(input_data_targettest):
    """
    Makes sure we raise a proper message to the user,
    when providing not supported data input
    """
    validator = TargetValidator()
    with pytest.raises(ValueError, match=r"legacy multi-.* data representation."):
        validator.fit(input_data_targettest)


def test_target_unsupported():
    """
    Makes sure we raise a proper message to the user,
    when providing not supported data input
    """
    validator = TargetValidator(is_classification=True)
    with pytest.raises(ValueError, match=r"The dimensionality of the train and test targets"):
        validator.fit(
            np.array([[0, 1, 0], [0, 1, 1]]),
            np.array([[0, 1, 0, 0], [0, 1, 1, 1]]),
        )
    with pytest.raises(ValueError, match=r"Train and test targets must both have the same dtypes"):
        validator.fit(
            pd.DataFrame({'a': [1, 2, 3]}),
            pd.DataFrame({'a': [True, False, False]}),
        )
    with pytest.raises(ValueError, match=r"Provided targets are not supported.*"):
        validator.fit(
            np.array([[0, 1, 2], [0, 3, 4]]),
            np.array([[0, 1, 2, 5], [0, 3, 4, 6]]),
        )
    with pytest.raises(ValueError, match="Train and test targets must both have the same"):
        validator.fit(
            pd.DataFrame({'string': ['foo']}),
            pd.DataFrame({'int': [1]}),
        )
    with pytest.raises(ValueError, match=r"Auto-sklearn only supports Numpy arrays, .*"):
        validator.fit({'input1': 1, 'input2': 2})
    with pytest.raises(ValueError, match=r"arget values cannot contain missing/NaN values"):
        validator.fit(np.array([np.nan, 1, 2]))
    with pytest.raises(ValueError, match=r"arget values cannot contain missing/NaN values"):
        validator.fit(sparse.csr_matrix(np.array([1, 2, np.nan])))
    with pytest.raises(ValueError, match=r"Cannot call transform on a validator that is not fit"):
        validator.transform(np.array([1, 2, 3]))
    with pytest.raises(ValueError, match=r"Cannot call inverse_transform on a validator that is"):
        validator.inverse_transform(np.array([1, 2, 3]))
    with pytest.raises(ValueError, match=r"Multi-dimensional classification is not yet supported"):
        validator._fit(np.array([[1, 2, 3], [1, 5, 6]]))

    # Dia/ DOK are not supported as type of target makes calls len on the array
    # which causes TypeError: len() of unsized object. Basically, sparse data as
    # multi-label is the only thing that makes sense in this format.
    with pytest.raises(ValueError, match=r"The provided data could not be interpreted by Sklearn"):
        validator.fit(sparse.dia_matrix(np.array([1, 2, 3])))

    validator.fit(np.array([[0, 1, 0], [0, 1, 1]]))
    with pytest.raises(ValueError, match=r"Number of outputs changed from"):
        validator.fit(np.array([0, 1, 0]))


def test_targetvalidator_inversetransform():
    """
    Test that the encoding/decoding works in 1D
    """
    validator = TargetValidator(is_classification=True)
    validator.fit(
        pd.DataFrame(data=['a', 'a', 'b', 'c', 'a'], dtype='category'),
    )
    y = validator.transform(
        pd.DataFrame(data=['a', 'a', 'b', 'c', 'a'], dtype='category'),
    )
    np.testing.assert_array_almost_equal(np.array([0, 0, 1, 2, 0]), y)

    y_decoded = validator.inverse_transform(y)
    assert ['a', 'a', 'b', 'c', 'a'] == y_decoded.tolist()

    assert validator.classes_.tolist() == ['a', 'b', 'c']

    validator = TargetValidator(is_classification=True)
    multi_label = pd.DataFrame(
        np.array([[1, 0, 0, 1], [0, 0, 1, 1], [0, 0, 0, 0]]),
        dtype=bool
    )
    validator.fit(multi_label)
    y = validator.transform(multi_label)

    y_decoded = validator.inverse_transform(y)
    np.testing.assert_array_almost_equal(y, y_decoded)

    # Multilabel classification is not encoded
    # For this reason, classes_ attribute does not contain a class
    np.testing.assert_array_almost_equal(validator.classes_, np.array([]))


# Actual checks for the targets
@pytest.mark.parametrize(
    'input_data_targettest',
    (
        'series_binary',
        'series_multiclass',
        'pandas_binary',
        'pandas_multiclass',
        'numpy_binary',
        'numpy_multiclass',
        'list_binary',
        'list_multiclass',
    ),
    indirect=True
)
def test_unknown_categories_in_targets(input_data_targettest):
    validator = TargetValidator(is_classification=True)
    validator.fit(input_data_targettest)

    # Add an extra category
    if isinstance(input_data_targettest, list):
        input_data_targettest.append(input_data_targettest[-1] + 5000)
    elif isinstance(input_data_targettest, (pd.DataFrame, pd.Series)):
        input_data_targettest.iloc[-1] = 5000
    elif isinstance(input_data_targettest, np.ndarray):
        input_data_targettest[-1] = 5000

    x_t = validator.transform(input_data_targettest)
    assert x_t[-1].item(0) == -1


def test_is_single_column_target():
    validator = TargetValidator(is_classification=True)
    validator.fit(np.array([1, 2, 3, 4]))
    assert validator.is_single_column_target()

    validator = TargetValidator(is_classification=True)
    validator.fit(np.array([[1, 0, 1, 0], [1, 1, 1, 1]]))
    assert not validator.is_single_column_target()



import os
import shutil
import unittest
import unittest.mock

import numpy as np

from autosklearn.automl_common.common.ensemble_building.abstract_ensemble import AbstractEnsemble

from autosklearn.metrics import make_scorer
from autosklearn.ensemble_builder import EnsembleBuilder


def scorer_function(a, b):
    return 0.9


MockMetric = make_scorer('mock', scorer_function)


class BackendMock(object):

    def __init__(self, target_directory):
        this_directory = os.path.abspath(
            os.path.dirname(__file__)
        )
        shutil.copytree(os.path.join(this_directory, 'data'), os.path.join(target_directory))
        self.temporary_directory = target_directory
        self.internals_directory = os.path.join(self.temporary_directory, '.auto-sklearn')

    def load_datamanager(self):
        manager = unittest.mock.Mock()
        manager.__reduce__ = lambda self: (unittest.mock.MagicMock, ())
        array = np.load(os.path.join(
            self.temporary_directory,
            '.auto-sklearn',
            'runs', '0_3_100.0',
            'predictions_test_0_3_100.0.npy'
        ))
        manager.data.get.return_value = array
        return manager

    def load_targets_ensemble(self):
        with open(os.path.join(
            self.temporary_directory,
            ".auto-sklearn",
            "predictions_ensemble_true.npy"
        ), "rb") as fp:
            y = np.load(fp, allow_pickle=True)
        return y

    def save_ensemble(self, ensemble, index_run, seed):
        return

    def save_predictions_as_txt(self, predictions, subset, idx, prefix, precision):
        return

    def get_runs_directory(self) -> str:
        return os.path.join(self.temporary_directory, '.auto-sklearn', 'runs')

    def get_numrun_directory(self, seed: int, num_run: int, budget: float) -> str:
        return os.path.join(self.get_runs_directory(), '%d_%d_%s' % (seed, num_run, budget))

    def get_model_filename(self, seed: int, idx: int, budget: float) -> str:
        return '%s.%s.%s.model' % (seed, idx, budget)


def compare_read_preds(read_preds1, read_preds2):
    """
    compares read_preds attribute. An alternative to
    assert Dict Equal as it contains np arrays, so we have
    to use np testing utilities accordingly
    """

    # Both arrays should have the same splits
    assert set(read_preds1.keys()) == set(read_preds2.keys())

    for k, v in read_preds1.items():

        # Each split should have the same elements
        assert set(read_preds1[k].keys()) == set(read_preds2[k].keys())

        # This level contains the scores/ensmebles/etc
        for actual_k, actual_v in read_preds1[k].items():

            # If it is a numpy array, make sure it is the same
            if type(actual_v) is np.ndarray:
                np.testing.assert_array_equal(actual_v, read_preds2[k][actual_k])
            else:
                assert actual_v == read_preds2[k][actual_k]


class EnsembleBuilderMemMock(EnsembleBuilder):

    def fit_ensemble(self, selected_keys):
        return True

    def predict(self, set_: str,
                ensemble: AbstractEnsemble,
                selected_keys: list,
                n_preds: int,
                index_run: int):
        np.ones([10000000, 1000000])



# -*- encoding: utf-8 -*-
import unittest

from autosklearn.pipeline.util import get_dataset
from autosklearn.classification import AutoSklearnClassifier

from autosklearn.smbo import _calculate_metafeatures, _calculate_metafeatures_encoded
from autosklearn.constants import REGRESSION, MULTICLASS_CLASSIFICATION
from autosklearn.metalearning.mismbo import suggest_via_metalearning
from autosklearn.util.pipeline import get_configuration_space
from sklearn.datasets import load_breast_cancer


class MetafeatureValueDummy(object):

    def __init__(self, name, value):
        self.name = name
        self.value = value


class Test(unittest.TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        self.X_train, self.Y_train, self.X_test, self.Y_test = \
            get_dataset('iris')

        eliminate_class_two = self.Y_train != 2
        self.X_train = self.X_train[eliminate_class_two]
        self.Y_train = self.Y_train[eliminate_class_two]

    @unittest.skip('TODO refactor!')
    def test_metalearning(self):
        dataset_name_classification = 'digits'
        initial_challengers_classification = {
            "ACC_METRIC": "--initial-challengers \" "
                          "-balancing:strategy 'weighting' "
                          "-classifier:__choice__ 'proj_logit'",
            "AUC_METRIC": "--initial-challengers \" "
                          "-balancing:strategy 'weighting' "
                          "-classifier:__choice__ 'liblinear_svc'",
            "BAC_METRIC": "--initial-challengers \" "
                          "-balancing:strategy 'weighting' "
                          "-classifier:__choice__ 'proj_logit'",
            "F1_METRIC": "--initial-challengers \" "
                         "-balancing:strategy 'weighting' "
                         "-classifier:__choice__ 'proj_logit'",
            "PAC_METRIC": "--initial-challengers \" "
                          "-balancing:strategy 'none' "
                          "-classifier:__choice__ 'random_forest'"
        }

        dataset_name_regression = 'diabetes'
        initial_challengers_regression = {
            "A_METRIC": "--initial-challengers \" "
                        "-imputation:strategy 'mean' "
                        "-one_hot_encoding:minimum_fraction '0.01' "
                        "-one_hot_encoding:use_minimum_fraction 'True' "
                        "-preprocessor:__choice__ 'no_preprocessing' "
                        "-regressor:__choice__ 'random_forest'",
            "R2_METRIC": "--initial-challengers \" "
                         "-imputation:strategy 'mean' "
                         "-one_hot_encoding:minimum_fraction '0.01' "
                         "-one_hot_encoding:use_minimum_fraction 'True' "
                         "-preprocessor:__choice__ 'no_preprocessing' "
                         "-regressor:__choice__ 'random_forest'",
        }

        for dataset_name, task, initial_challengers in [
            (dataset_name_regression, REGRESSION, initial_challengers_regression),
            (dataset_name_classification, MULTICLASS_CLASSIFICATION,
             initial_challengers_classification)]:

            for metric in initial_challengers:
                configuration_space = get_configuration_space(
                    {
                        'metric': metric,
                        'task': task,
                        'is_sparse': False
                    },
                    include={'feature_preprocessor': ['no_preprocessing']})

                X_train, Y_train, X_test, Y_test = get_dataset(dataset_name)
                categorical = {i: False for i in range(X_train.shape[1])}

                meta_features_label = _calculate_metafeatures(
                    X_train, Y_train, categorical, dataset_name, task)
                meta_features_encoded_label = _calculate_metafeatures_encoded(
                    X_train, Y_train, categorical, dataset_name, task)

                initial_configuration_strings_for_smac = \
                    suggest_via_metalearning(
                        meta_features_label,
                        meta_features_encoded_label,
                        configuration_space, dataset_name, metric,
                        task, False, 1, None)

                print(metric)
                print(initial_configuration_strings_for_smac[0])
                self.assertTrue(initial_configuration_strings_for_smac[
                                    0].startswith(initial_challengers[metric]))

    def test_metadata_directory(self):
        # Test that metadata directory is set correctly (if user specifies,
        # Auto-sklearn should check that the directory exists. If not, it
        # should use the default directory.
        dask_client = unittest.mock.Mock()
        automl1 = AutoSklearnClassifier(
            time_left_for_this_task=30,
            per_run_time_limit=5,
            metadata_directory="pyMetaLearn/metadata_dir",  # user specified metadata_dir
            dask_client=dask_client,
        )
        self.assertEqual(automl1.metadata_directory,
                         "pyMetaLearn/metadata_dir")

        automl2 = AutoSklearnClassifier(  # default metadata_dir
            time_left_for_this_task=30,
            per_run_time_limit=5,
            dask_client=dask_client,
        )
        self.assertIsNone(automl2.metadata_directory)

        nonexistent_dir = "nonexistent_dir"
        automl3 = AutoSklearnClassifier(
            time_left_for_this_task=30,
            per_run_time_limit=5,
            metadata_directory=nonexistent_dir,  # user specified metadata_dir
            dask_client=dask_client,
            ensemble_size=0,
        )
        X, y = load_breast_cancer(return_X_y=True)
        self.assertRaisesRegex(ValueError, "The specified metadata directory "
                               "\'%s\' does not exist!" % nonexistent_dir,
                               automl3.fit, X=X, y=y)



from collections import OrderedDict

import unittest
import numpy

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter

from autosklearn.pipeline.components.classification.liblinear_svc import LibLinear_SVC
from autosklearn.pipeline.components.classification.lda import LDA

from autosklearn.pipeline.components.feature_preprocessing.pca import PCA
from autosklearn.pipeline.components.feature_preprocessing.truncatedSVD import TruncatedSVD
from autosklearn.pipeline.components.feature_preprocessing.no_preprocessing import NoPreprocessing
from autosklearn.pipeline.components.feature_preprocessing.random_trees_embedding \
    import RandomTreesEmbedding
import autosklearn.pipeline.create_searchspace_util


class TestCreateClassificationSearchspace(unittest.TestCase):
    _multiprocess_can_split_ = True

    def test_get_match_array_sparse_and_dense(self):
        # preproc is empty
        preprocessors = OrderedDict()
        preprocessors['pca'] = PCA
        classifiers = OrderedDict()
        classifiers['lda'] = LDA
        # Sparse + dense

        class Preprocessors(object):
            @classmethod
            def get_available_components(self, *args, **kwargs):
                return preprocessors

        class Classifiers(object):
            @classmethod
            def get_available_components(self, *args, **kwargs):
                return classifiers

        # Dense
        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, PCA), (1, LDA)), dataset_properties={'sparse': True})
        self.assertEqual(numpy.sum(m), 0)

        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, PCA), (1, LDA)), dataset_properties={'sparse': False})
        self.assertEqual(m, [[1]])

        # Sparse
        preprocessors['tSVD'] = TruncatedSVD
        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, LDA)),
            dataset_properties={'sparse': True})
        self.assertEqual(m[0], [0])  # pca
        self.assertEqual(m[1], [1])  # svd

        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, LDA)),
            dataset_properties={'sparse': False})
        self.assertEqual(m[0], [1])  # pca
        self.assertEqual(m[1], [0])  # svd

        preprocessors['none'] = NoPreprocessing
        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, LDA)),
            dataset_properties={'sparse': True})
        self.assertEqual(m[0, :], [0])  # pca
        self.assertEqual(m[1, :], [1])  # tsvd
        self.assertEqual(m[2, :], [0])  # none

        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, LDA)),
            dataset_properties={'sparse': False})
        self.assertEqual(m[0, :], [1])  # pca
        self.assertEqual(m[1, :], [0])  # tsvd
        self.assertEqual(m[2, :], [1])  # none

        classifiers['libsvm'] = LibLinear_SVC
        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, Classifiers)),
            dataset_properties={'sparse': False})
        self.assertListEqual(list(m[0, :]), [1, 1])  # pca
        self.assertListEqual(list(m[1, :]), [0, 0])  # tsvd
        self.assertListEqual(list(m[2, :]), [1, 1])  # none

        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, Classifiers)),
            dataset_properties={'sparse': True})
        self.assertListEqual(list(m[0, :]), [0, 0])  # pca
        self.assertListEqual(list(m[1, :]), [1, 1])  # tsvd
        self.assertListEqual(list(m[2, :]), [0, 1])  # none

        # Do fancy 3d stuff
        preprocessors['random_trees'] = RandomTreesEmbedding
        m = autosklearn.pipeline.create_searchspace_util.get_match_array(
            pipeline=((0, Preprocessors), (1, Preprocessors), (2, Classifiers)),
            dataset_properties={'sparse': False})
        # PCA followed by truncated SVD is forbidden
        self.assertEqual(list(m[0].flatten()), [1, 1, 0, 0, 1, 1, 0, 1])
        # Truncated SVD is forbidden
        self.assertEqual(list(m[1].flatten()), [0, 0, 0, 0, 0, 0, 0, 0])
        # Truncated SVD is forbidden after no_preprocessing
        self.assertEqual(list(m[2].flatten()), [1, 1, 0, 0, 1, 1, 0, 1])
        # PCA is forbidden, truncatedSVD allowed after random trees embedding
        # lda only allowed after truncatedSVD
        self.assertEqual(list(m[3].flatten()), [0, 0, 1, 1, 0, 1, 0, 1])

    def test_get_match_array_signed_unsigned_and_binary(self):
        pass

    @unittest.skip("Not currently working.")
    def test_add_forbidden(self):
        m = numpy.ones([2, 3])
        preprocessors_list = ['pa', 'pb']
        classifier_list = ['ca', 'cb', 'cc']
        cs = ConfigurationSpace()
        preprocessor = CategoricalHyperparameter(name='feature_preprocessor',
                                                 choices=preprocessors_list)
        classifier = CategoricalHyperparameter(name='classifier',
                                               choices=classifier_list)
        cs.add_hyperparameter(preprocessor)
        cs.add_hyperparameter(classifier)
        new_cs = autosklearn.pipeline.create_searchspace_util.add_forbidden(
            conf_space=cs, node_0_list=preprocessors_list,
            node_1_list=classifier_list, matches=m,
            node_0_name='feature_preprocessor', node_1_name="classifier")
        self.assertEqual(len(new_cs.forbidden_clauses), 0)
        self.assertIsInstance(new_cs, ConfigurationSpace)

        m[1, 1] = 0
        new_cs = autosklearn.pipeline.create_searchspace_util.add_forbidden(
            conf_space=cs, node_0_list=preprocessors_list,
            node_1_list=classifier_list, matches=m,
            node_0_name='feature_preprocessor', node_1_name="classifier")
        self.assertEqual(len(new_cs.forbidden_clauses), 1)
        self.assertEqual(new_cs.forbidden_clauses[0].components[0].value, 'cb')
        self.assertEqual(new_cs.forbidden_clauses[0].components[1].value, 'pb')
        self.assertIsInstance(new_cs, ConfigurationSpace)



# -*- encoding: utf-8 -*-
import itertools
import os
import pickle
import sys
import time
import glob
import unittest
import unittest.mock
import warnings

import numpy as np
import pandas as pd
import pytest
from scipy.sparse import csr_matrix
import sklearn.datasets
from sklearn.ensemble import VotingRegressor, VotingClassifier
from smac.scenario.scenario import Scenario
from smac.facade.roar_facade import ROAR

from autosklearn.automl import AutoML, AutoMLClassifier, AutoMLRegressor, _model_predict
from autosklearn.data.validation import InputValidator
import autosklearn.automl
from autosklearn.data.xy_data_manager import XYDataManager
from autosklearn.metrics import (
    accuracy, log_loss, balanced_accuracy, default_metric_for_task
)
from autosklearn.evaluation.abstract_evaluator import MyDummyClassifier, MyDummyRegressor
from autosklearn.util.logging_ import PickableLoggerAdapter
import autosklearn.pipeline.util as putil
from autosklearn.constants import (
    MULTICLASS_CLASSIFICATION,
    BINARY_CLASSIFICATION,
    MULTILABEL_CLASSIFICATION,
    REGRESSION,
    MULTIOUTPUT_REGRESSION,
    CLASSIFICATION_TASKS,
    REGRESSION_TASKS,
)
from smac.tae import StatusType

sys.path.append(os.path.dirname(__file__))
from automl_utils import print_debug_information, count_succeses, AutoMLLogParser, includes_all_scores, includes_train_scores, performance_over_time_is_plausible  # noqa (E402: module level import not at top of file)


class AutoMLStub(AutoML):
    def __init__(self):
        self.__class__ = AutoML
        self._task = None
        self._dask_client = None
        self._is_dask_client_internally_created = False

    def __del__(self):
        pass


def test_fit(dask_client):

    X_train, Y_train, X_test, Y_test = putil.get_dataset('iris')
    automl = autosklearn.automl.AutoML(
        seed=0,
        time_left_for_this_task=30,
        per_run_time_limit=5,
        metric=accuracy,
        dask_client=dask_client,
    )

    automl.fit(X_train, Y_train, task=MULTICLASS_CLASSIFICATION)

    score = automl.score(X_test, Y_test)
    assert score > 0.8
    assert count_succeses(automl.cv_results_) > 0
    assert includes_train_scores(automl.performance_over_time_.columns) is True
    assert performance_over_time_is_plausible(automl.performance_over_time_) is True
    assert automl._task == MULTICLASS_CLASSIFICATION

    del automl


def test_fit_roar(dask_client_single_worker):
    def get_roar_object_callback(
            scenario_dict,
            seed,
            ta,
            ta_kwargs,
            dask_client,
            n_jobs,
            **kwargs
    ):
        """Random online adaptive racing.

        http://ml.informatik.uni-freiburg.de/papers/11-LION5-SMAC.pdf"""
        scenario = Scenario(scenario_dict)
        return ROAR(
            scenario=scenario,
            rng=seed,
            tae_runner=ta,
            tae_runner_kwargs=ta_kwargs,
            dask_client=dask_client,
            n_jobs=n_jobs,
        )

    X_train, Y_train, X_test, Y_test = putil.get_dataset('iris')
    automl = autosklearn.automl.AutoML(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        initial_configurations_via_metalearning=0,
        get_smac_object_callback=get_roar_object_callback,
        metric=accuracy,
        dask_client=dask_client_single_worker,
    )

    automl.fit(X_train, Y_train, task=MULTICLASS_CLASSIFICATION)

    score = automl.score(X_test, Y_test)
    assert score > 0.8
    assert count_succeses(automl.cv_results_) > 0
    assert includes_train_scores(automl.performance_over_time_.columns) is True
    assert automl._task == MULTICLASS_CLASSIFICATION

    del automl


def test_refit_shuffle_on_fail(dask_client):

    failing_model = unittest.mock.Mock()
    failing_model.fit.side_effect = [ValueError(), ValueError(), None]
    failing_model.fit_transformer.side_effect = [
        ValueError(), ValueError(), (None, {})]
    failing_model.get_max_iter.return_value = 100

    auto = AutoML(30, 5, dask_client=dask_client)
    ensemble_mock = unittest.mock.Mock()
    ensemble_mock.get_selected_model_identifiers.return_value = [(1, 1, 50.0)]
    auto.ensemble_ = ensemble_mock
    auto.InputValidator = InputValidator()
    for budget_type in [None, 'iterations']:
        auto._budget_type = budget_type

        auto.models_ = {(1, 1, 50.0): failing_model}

        # Make sure a valid 2D array is given to automl
        X = np.array([1, 2, 3]).reshape(-1, 1)
        y = np.array([1, 2, 3])
        auto.InputValidator.fit(X, y)
        auto.refit(X, y)

        assert failing_model.fit.call_count == 3
    assert failing_model.fit_transformer.call_count == 3

    del auto


def test_only_loads_ensemble_models(automl_stub):

    def side_effect(ids, *args, **kwargs):
        return models if ids is identifiers else {}

    # Add a resampling strategy as this is required by load_models
    automl_stub._resampling_strategy = 'holdout'
    identifiers = [(1, 2), (3, 4)]

    models = [42]
    load_ensemble_mock = unittest.mock.Mock()
    load_ensemble_mock.get_selected_model_identifiers.return_value = identifiers
    automl_stub._backend.load_ensemble.return_value = load_ensemble_mock
    automl_stub._backend.load_models_by_identifiers.side_effect = side_effect

    automl_stub._load_models()
    assert models == automl_stub.models_
    assert automl_stub.cv_models_ is None

    automl_stub._resampling_strategy = 'cv'

    models = [42]
    automl_stub._backend.load_cv_models_by_identifiers.side_effect = side_effect

    automl_stub._load_models()
    assert models == automl_stub.cv_models_


def test_check_for_models_if_no_ensemble(automl_stub):
    models = [42]
    automl_stub._backend.load_ensemble.return_value = None
    automl_stub._backend.list_all_models.return_value = models
    automl_stub._disable_evaluator_output = False

    automl_stub._load_models()


def test_raises_if_no_models(automl_stub):
    automl_stub._backend.load_ensemble.return_value = None
    automl_stub._backend.list_all_models.return_value = []
    automl_stub._resampling_strategy = 'holdout'

    automl_stub._disable_evaluator_output = False
    with pytest.raises(ValueError):
        automl_stub._load_models()

    automl_stub._disable_evaluator_output = True
    automl_stub._load_models()


def test_delete_non_candidate_models(dask_client):

    seed = 555
    X, Y, _, _ = putil.get_dataset('iris')
    automl = autosklearn.automl.AutoML(
        delete_tmp_folder_after_terminate=False,
        time_left_for_this_task=60,
        per_run_time_limit=5,
        ensemble_nbest=3,
        seed=seed,
        initial_configurations_via_metalearning=0,
        resampling_strategy='holdout',
        include={
            'classifier': ['sgd'],
            'feature_preprocessor': ['no_preprocessing']
        },
        metric=accuracy,
        dask_client=dask_client,
        # Force model to be deleted. That is, from 50 which is the
        # default to 3 to make sure we delete models.
        max_models_on_disc=3,
    )

    automl.fit(X, Y, task=MULTICLASS_CLASSIFICATION, X_test=X, y_test=Y)

    # Assert at least one model file has been deleted and that there were no
    # deletion errors
    log_file_path = glob.glob(os.path.join(
        automl._backend.temporary_directory, 'AutoML(' + str(seed) + '):*.log'))
    with open(log_file_path[0]) as log_file:
        log_content = log_file.read()
        assert 'Deleted files of non-candidate model' in log_content, log_content
        assert 'Failed to delete files of non-candidate model' not in log_content, log_content
        assert 'Failed to lock model' not in log_content, log_content

    # Assert that the files of the models used by the ensemble weren't deleted
    model_files = automl._backend.list_all_models(seed=seed)
    model_files_idx = set()
    for m_file in model_files:
        # Extract the model identifiers from the filename
        m_file = os.path.split(m_file)[1].replace('.model', '').split('.', 2)
        model_files_idx.add((int(m_file[0]), int(m_file[1]), float(m_file[2])))
    ensemble_members_idx = set(automl.ensemble_.identifiers_)
    assert ensemble_members_idx.issubset(model_files_idx), (ensemble_members_idx, model_files_idx)

    del automl


def test_binary_score_and_include(dask_client):
    """
    Test fix for binary classification prediction
    taking the index 1 of second dimension in prediction matrix
    """

    data = sklearn.datasets.make_classification(
        n_samples=400, n_features=10, n_redundant=1, n_informative=3,
        n_repeated=1, n_clusters_per_class=2, random_state=1)
    X_train = data[0][:200]
    Y_train = data[1][:200]
    X_test = data[0][200:]
    Y_test = data[1][200:]

    automl = autosklearn.automl.AutoML(
        20, 5,
        include={'classifier': ['sgd'],
                 'feature_preprocessor': ['no_preprocessing']},
        metric=accuracy,
        dask_client=dask_client,
    )

    automl.fit(X_train, Y_train, task=BINARY_CLASSIFICATION)

    assert automl._task == BINARY_CLASSIFICATION

    # TODO, the assumption from above is not really tested here
    # Also, the score method should be removed, it only makes little sense
    score = automl.score(X_test, Y_test)
    assert score >= 0.4

    del automl


def test_automl_outputs(dask_client):

    X_train, Y_train, X_test, Y_test = putil.get_dataset('iris')
    name = 'iris'
    auto = autosklearn.automl.AutoML(
        30, 5,
        initial_configurations_via_metalearning=0,
        seed=100,
        metric=accuracy,
        dask_client=dask_client,
        delete_tmp_folder_after_terminate=False,
    )

    auto.fit(
        X=X_train,
        y=Y_train,
        X_test=X_test,
        y_test=Y_test,
        dataset_name=name,
        task=MULTICLASS_CLASSIFICATION,
    )

    data_manager_file = os.path.join(
        auto._backend.temporary_directory,
        '.auto-sklearn',
        'datamanager.pkl'
    )

    # pickled data manager (without one hot encoding!)
    with open(data_manager_file, 'rb') as fh:
        D = pickle.load(fh)
        assert np.allclose(D.data['X_train'], X_train)

    # Check that all directories are there
    fixture = [
        'true_targets_ensemble.npy',
        'start_time_100',
        'datamanager.pkl',
        'ensemble_read_preds.pkl',
        'ensemble_read_losses.pkl',
        'runs',
        'ensembles',
        'ensemble_history.json',
    ]
    assert (
        sorted(os.listdir(os.path.join(auto._backend.temporary_directory,
                                       '.auto-sklearn')))
        == sorted(fixture)
    )

    # At least one ensemble, one validation, one test prediction and one
    # model and one ensemble
    fixture = glob.glob(os.path.join(
        auto._backend.temporary_directory,
        '.auto-sklearn', 'runs', '*', 'predictions_ensemble*npy',
    ))
    assert len(fixture) > 0

    fixture = glob.glob(os.path.join(auto._backend.temporary_directory, '.auto-sklearn',
                                     'runs', '*', '100.*.model'))
    assert len(fixture) > 0

    fixture = os.listdir(os.path.join(auto._backend.temporary_directory,
                                      '.auto-sklearn', 'ensembles'))
    assert '100.0000000000.ensemble' in fixture

    # Start time
    start_time_file_path = os.path.join(auto._backend.temporary_directory,
                                        '.auto-sklearn', "start_time_100")
    with open(start_time_file_path, 'r') as fh:
        start_time = float(fh.read())
    assert time.time() - start_time >= 10, print_debug_information(auto)

    # Then check that the logger matches the run expectation
    logfile = glob.glob(os.path.join(
           auto._backend.temporary_directory, 'AutoML*.log'))[0]
    parser = AutoMLLogParser(logfile)

    # The number of ensemble trajectories properly in log file
    success_ensemble_iters_auto = len(auto.ensemble_performance_history)
    success_ensemble_iters_log = parser.count_ensembler_success_pynisher_calls()
    assert success_ensemble_iters_auto == success_ensemble_iters_log, "{} != {}".format(
        auto.ensemble_performance_history,
        print_debug_information(auto),
    )

    # We also care that no iteration got lost
    # This is important because it counts for pynisher calls
    # and whether a pynisher call actually called the ensemble
    total_ensemble_iterations = parser.count_ensembler_iterations()
    assert len(total_ensemble_iterations) > 1  # At least 1 iteration
    assert range(1, max(total_ensemble_iterations) + 1), total_ensemble_iterations

    # a point where pynisher is called before budget exhaustion
    # Dummy not in run history
    total_calls_to_pynisher_log = parser.count_tae_pynisher_calls() - 1
    total_returns_from_pynisher_log = parser.count_tae_pynisher_returns() - 1
    total_elements_rh = len([run_value for run_value in auto.runhistory_.data.values(
    ) if run_value.status == StatusType.RUNNING])

    # Make sure we register all calls to pynisher
    # The less than or equal here is added as a WA as
    # https://github.com/automl/SMAC3/pull/712 is not yet integrated
    assert total_elements_rh <= total_calls_to_pynisher_log, print_debug_information(auto)

    # Make sure we register all returns from pynisher
    assert total_elements_rh <= total_returns_from_pynisher_log, print_debug_information(auto)

    # Lastly check that settings are print to logfile
    ensemble_size = parser.get_automl_setting_from_log(auto._dataset_name, 'ensemble_size')
    assert auto._ensemble_size == int(ensemble_size)

    del auto


@pytest.mark.parametrize("datasets", [('breast_cancer', BINARY_CLASSIFICATION),
                                      ('wine', MULTICLASS_CLASSIFICATION),
                                      ('diabetes', REGRESSION)])
def test_do_dummy_prediction(dask_client, datasets):

    name, task = datasets

    X_train, Y_train, X_test, Y_test = putil.get_dataset(name)
    datamanager = XYDataManager(
        X_train, Y_train,
        X_test, Y_test,
        task=task,
        dataset_name=name,
        feat_type={i: 'numerical' for i in range(X_train.shape[1])},
    )

    auto = autosklearn.automl.AutoML(
        20, 5,
        initial_configurations_via_metalearning=25,
        metric=accuracy,
        dask_client=dask_client,
        delete_tmp_folder_after_terminate=False,
    )
    auto._backend = auto._create_backend()

    # Make a dummy logger
    auto._logger_port = 9020
    auto._logger = unittest.mock.Mock()
    auto._logger.info.return_value = None

    auto._backend.save_datamanager(datamanager)
    D = auto._backend.load_datamanager()

    # Check if data manager is correcly loaded
    assert D.info['task'] == datamanager.info['task']
    auto._do_dummy_prediction(D, 1)

    # Ensure that the dummy predictions are not in the current working
    # directory, but in the temporary directory.
    unexpected_directory = os.path.join(os.getcwd(), '.auto-sklearn')
    expected_directory = os.path.join(
        auto._backend.temporary_directory,
        '.auto-sklearn',
        'runs',
        '1_1_0.0',
        'predictions_ensemble_1_1_0.0.npy'
    )
    assert not os.path.exists(unexpected_directory)
    assert os.path.exists(expected_directory)

    auto._clean_logger()

    del auto


@unittest.mock.patch('autosklearn.evaluation.ExecuteTaFuncWithQueue.run')
def test_fail_if_dummy_prediction_fails(ta_run_mock, dask_client):

    X_train, Y_train, X_test, Y_test = putil.get_dataset('iris')
    datamanager = XYDataManager(
        X_train, Y_train,
        X_test, Y_test,
        task=2,
        feat_type={i: 'Numerical' for i in range(X_train.shape[1])},
        dataset_name='iris',
    )

    time_for_this_task = 30
    per_run_time = 10
    auto = autosklearn.automl.AutoML(time_for_this_task,
                                     per_run_time,
                                     initial_configurations_via_metalearning=25,
                                     metric=accuracy,
                                     dask_client=dask_client,
                                     delete_tmp_folder_after_terminate=False,
                                     )
    auto._backend = auto._create_backend()
    auto._backend._make_internals_directory()
    auto._backend.save_datamanager(datamanager)

    # Make a dummy logger
    auto._logger_port = 9020
    auto._logger = unittest.mock.Mock()
    auto._logger.info.return_value = None

    # First of all, check that ta.run() is actually called.
    ta_run_mock.return_value = StatusType.SUCCESS, None, None, {}
    auto._do_dummy_prediction(datamanager, 1)
    ta_run_mock.assert_called_once_with(1, cutoff=time_for_this_task)

    # Case 1. Check that function raises no error when statustype == success.
    # ta.run() returns status, cost, runtime, and additional info.
    ta_run_mock.return_value = StatusType.SUCCESS, None, None, {}
    raised = False
    try:
        auto._do_dummy_prediction(datamanager, 1)
    except ValueError:
        raised = True
    assert not raised, 'Exception raised'

    # Case 2. Check that if statustype returned by ta.run() != success,
    # the function raises error.
    ta_run_mock.return_value = StatusType.CRASHED, None, None, {}
    with pytest.raises(
        ValueError,
        match='Dummy prediction failed with run state StatusType.CRASHED and additional output: {}.'  # noqa
    ):
        auto._do_dummy_prediction(datamanager, 1)

    ta_run_mock.return_value = StatusType.ABORT, None, None, {}
    with pytest.raises(
        ValueError,
        match='Dummy prediction failed with run state StatusType.ABORT '
              'and additional output: {}.',
    ):
        auto._do_dummy_prediction(datamanager, 1)
    ta_run_mock.return_value = StatusType.TIMEOUT, None, None, {}
    with pytest.raises(
        ValueError,
        match='Dummy prediction failed with run state StatusType.TIMEOUT '
              'and additional output: {}.'
    ):
        auto._do_dummy_prediction(datamanager, 1)
    ta_run_mock.return_value = StatusType.MEMOUT, None, None, {}
    with pytest.raises(
        ValueError,
        match='Dummy prediction failed with run state StatusType.MEMOUT '
              'and additional output: {}.',
    ):
        auto._do_dummy_prediction(datamanager, 1)
    ta_run_mock.return_value = StatusType.CAPPED, None, None, {}
    with pytest.raises(
        ValueError,
        match='Dummy prediction failed with run state StatusType.CAPPED '
              'and additional output: {}.'
    ):
        auto._do_dummy_prediction(datamanager, 1)

    ta_run_mock.return_value = StatusType.CRASHED, None, None, {'exitcode': -6}
    with pytest.raises(
        ValueError,
        match='The error suggests that the provided memory limits were too tight.',
    ):
        auto._do_dummy_prediction(datamanager, 1)


@unittest.mock.patch('autosklearn.smbo.AutoMLSMBO.run_smbo')
def test_exceptions_inside_log_in_smbo(smbo_run_mock, dask_client):

    # Below importing and shutdown is a workaround, to make sure
    # we reset the port to collect messages. Randomly, when running
    # this test with multiple other test at the same time causes this
    # test to fail. This resets the singletons of the logging class
    import logging
    logging.shutdown()

    automl = autosklearn.automl.AutoML(
        20,
        5,
        metric=accuracy,
        dask_client=dask_client,
        delete_tmp_folder_after_terminate=False,
    )

    dataset_name = 'test_exceptions_inside_log'

    # Create a custom exception to prevent other errors to slip in
    class MyException(Exception):
        pass

    X_train, Y_train, X_test, Y_test = putil.get_dataset('iris')
    # The first call is on dummy predictor failure
    message = str(np.random.randint(100)) + '_run_smbo'
    smbo_run_mock.side_effect = MyException(message)

    with pytest.raises(MyException):
        automl.fit(
            X_train,
            Y_train,
            task=MULTICLASS_CLASSIFICATION,
            dataset_name=dataset_name,
        )

    # make sure that the logfile was created
    logger_name = 'AutoML(%d):%s' % (1, dataset_name)
    logger = logging.getLogger(logger_name)
    logfile = os.path.join(automl._backend.temporary_directory, logger_name + '.log')
    assert os.path.exists(logfile), print_debug_information(automl) + str(automl._clean_logger())

    # Give some time for the error message to be printed in the
    # log file
    found_message = False
    for incr_tolerance in range(5):
        with open(logfile) as f:
            lines = f.readlines()
        if any(message in line for line in lines):
            found_message = True
            break
        else:
            time.sleep(incr_tolerance)

    # Speed up the closing after forced crash
    automl._clean_logger()

    if not found_message:
        pytest.fail("Did not find {} in the log file {} for logger {}/{}/{}".format(
            message,
            print_debug_information(automl),
            vars(automl._logger.logger),
            vars(logger),
            vars(logging.getLogger())
        ))


@pytest.mark.parametrize("metric", [log_loss, balanced_accuracy])
def test_load_best_individual_model(metric, dask_client):

    X_train, Y_train, X_test, Y_test = putil.get_dataset('iris')
    automl = autosklearn.automl.AutoML(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        metric=metric,
        dask_client=dask_client,
        delete_tmp_folder_after_terminate=False,
    )

    # We cannot easily mock a function sent to dask
    # so for this test we create the whole set of models/ensembles
    # but prevent it to be loaded
    automl.fit(X_train, Y_train, task=MULTICLASS_CLASSIFICATION)

    automl._backend.load_ensemble = unittest.mock.MagicMock(return_value=None)

    # A memory error occurs in the ensemble construction
    assert automl._backend.load_ensemble(automl._seed) is None

    # The load model is robust to this and loads the best model
    automl._load_models()
    assert automl.ensemble_ is not None

    # Just 1 model is there for ensemble and all weight must be on it
    get_models_with_weights = automl.get_models_with_weights()
    assert len(get_models_with_weights) == 1
    assert get_models_with_weights[0][0] == 1.0

    # Match a toy dataset
    if metric.name == 'balanced_accuracy':
        assert automl.score(X_test, Y_test) > 0.9
    elif metric.name == 'log_loss':
        # Seen values in github actions of 0.6978304740364537
        assert automl.score(X_test, Y_test) < 0.7
    else:
        raise ValueError(metric.name)

    del automl


def test_fail_if_feat_type_on_pandas_input(dask_client):
    """We do not support feat type when pandas
    is provided as an input
    """
    automl = autosklearn.automl.AutoML(
        time_left_for_this_task=30,
        per_run_time_limit=5,
        metric=accuracy,
        dask_client=dask_client,
    )

    X_train = pd.DataFrame({'a': [1, 1], 'c': [1, 2]})
    y_train = [1, 0]
    with pytest.raises(
        ValueError,
        match=""
        "providing the option feat_type to the fit method is not supported when using a Dataframe"
    ):
        automl.fit(
            X_train, y_train,
            task=BINARY_CLASSIFICATION,
            feat_type={1: 'Categorical', 2: 'Numerical'},
        )


@pytest.mark.parametrize(
    'memory_limit,precision,task',
    [
        (memory_limit, precision, task)
        for task in itertools.chain(CLASSIFICATION_TASKS, REGRESSION_TASKS)
        for precision in (float, np.float32, np.float64, np.float128)
        for memory_limit in (1, 100, None)
    ]
)
def test_subsample_if_too_large(memory_limit, precision, task):
    fixture = {
        BINARY_CLASSIFICATION: {
            1: {float: 1310, np.float32: 2621, np.float64: 1310, np.float128: 655},
            100: {float: 12000, np.float32: 12000, np.float64: 12000, np.float128: 12000},
            None: {float: 12000, np.float32: 12000, np.float64: 12000, np.float128: 12000},
        },
        MULTICLASS_CLASSIFICATION: {
            1: {float: 204, np.float32: 409, np.float64: 204, np.float128: 102},
            100: {float: 1797, np.float32: 1797, np.float64: 1797, np.float128: 1797},
            None: {float: 1797, np.float32: 1797, np.float64: 1797, np.float128: 1797},
        },
        MULTILABEL_CLASSIFICATION: {
            1: {float: 204, np.float32: 409, np.float64: 204, np.float128: 102},
            100: {float: 1797, np.float32: 1797, np.float64: 1797, np.float128: 1797},
            None: {float: 1797, np.float32: 1797, np.float64: 1797, np.float128: 1797},
        },
        REGRESSION: {
            1: {float: 655, np.float32: 1310, np.float64: 655, np.float128: 327},
            100: {float: 5000, np.float32: 5000, np.float64: 5000, np.float128: 5000},
            None: {float: 5000, np.float32: 5000, np.float64: 5000, np.float128: 5000},
        },
        MULTIOUTPUT_REGRESSION: {
            1: {float: 655, np.float32: 1310, np.float64: 655, np.float128: 327},
            100: {float: 5000, np.float32: 5000, np.float64: 5000, np.float128: 5000},
            None: {float: 5000, np.float32: 5000, np.float64: 5000, np.float128: 5000},
        }
    }
    mock = unittest.mock.Mock()
    if task == BINARY_CLASSIFICATION:
        X, y = sklearn.datasets.make_hastie_10_2()
    elif task == MULTICLASS_CLASSIFICATION:
        X, y = sklearn.datasets.load_digits(return_X_y=True)
    elif task == MULTILABEL_CLASSIFICATION:
        X, y_ = sklearn.datasets.load_digits(return_X_y=True)
        y = np.zeros((X.shape[0], 10))
        for i, j in enumerate(y_):
            y[i, j] = 1
    elif task == REGRESSION:
        X, y = sklearn.datasets.make_friedman1(n_samples=5000, n_features=20)
    elif task == MULTIOUTPUT_REGRESSION:
        X, y = sklearn.datasets.make_friedman1(n_samples=5000, n_features=20)
        y = np.vstack((y, y)).transpose()
    else:
        raise ValueError(task)
    X = X.astype(precision)

    assert X.shape[0] == y.shape[0]

    X_new, y_new = AutoML.subsample_if_too_large(X, y, mock, 1, memory_limit, task)
    assert X_new.shape[0] == fixture[task][memory_limit][precision]
    if memory_limit == 1:
        if precision in (np.float128, np.float64, float):
            assert mock.warning.call_count == 2
        else:
            assert mock.warning.call_count == 1
    else:
        assert mock.warning.call_count == 0


def data_input_and_target_types():
    n_rows = 100

    # Create valid inputs
    X_ndarray = np.random.random(size=(n_rows, 5))
    X_ndarray[X_ndarray < .9] = 0

    # Binary Classificaiton
    y_binary_ndarray = np.random.random(size=n_rows)
    y_binary_ndarray[y_binary_ndarray >= 0.5] = 1
    y_binary_ndarray[y_binary_ndarray < 0.5] = 0

    # Multiclass classification
    y_multiclass_ndarray = np.random.random(size=n_rows)
    y_multiclass_ndarray[y_multiclass_ndarray > 0.66] = 2
    y_multiclass_ndarray[(y_multiclass_ndarray <= 0.66) & (y_multiclass_ndarray >= 0.33)] = 1
    y_multiclass_ndarray[y_multiclass_ndarray < 0.33] = 0

    # Multilabel classificaiton
    y_multilabel_ndarray = np.random.random(size=(n_rows, 3))
    y_multilabel_ndarray[y_multilabel_ndarray > 0.5] = 1
    y_multilabel_ndarray[y_multilabel_ndarray <= 0.5] = 0

    # Regression
    y_regression_ndarray = np.random.random(size=n_rows)

    # Multioutput Regression
    y_multioutput_regression_ndarray = np.random.random(size=(n_rows, 3))

    xs = [
        X_ndarray,
        list(X_ndarray),
        csr_matrix(X_ndarray),
        pd.DataFrame(data=X_ndarray),
    ]

    ys_binary = [
        y_binary_ndarray,
        list(y_binary_ndarray),
        csr_matrix(y_binary_ndarray),
        pd.Series(y_binary_ndarray),
        pd.DataFrame(data=y_binary_ndarray),
    ]

    ys_multiclass = [
        y_multiclass_ndarray,
        list(y_multiclass_ndarray),
        csr_matrix(y_multiclass_ndarray),
        pd.Series(y_multiclass_ndarray),
        pd.DataFrame(data=y_multiclass_ndarray),
    ]

    ys_multilabel = [
        y_multilabel_ndarray,
        list(y_multilabel_ndarray),
        csr_matrix(y_multilabel_ndarray),
        # pd.Series(y_multilabel_ndarray)
        pd.DataFrame(data=y_multilabel_ndarray),
    ]

    ys_regression = [
        y_regression_ndarray,
        list(y_regression_ndarray),
        csr_matrix(y_regression_ndarray),
        pd.Series(y_regression_ndarray),
        pd.DataFrame(data=y_regression_ndarray),
    ]

    ys_multioutput_regression = [
        y_multioutput_regression_ndarray,
        list(y_multioutput_regression_ndarray),
        csr_matrix(y_multioutput_regression_ndarray),
        # pd.Series(y_multioutput_regression_ndarray),
        pd.DataFrame(data=y_multioutput_regression_ndarray),
    ]

    # [ (X, y, X_test, y_test, task), ... ]
    return (
        (X, y, X, y, task)
        for X in xs
        for y, task in itertools.chain(
            itertools.product(ys_binary, [BINARY_CLASSIFICATION]),
            itertools.product(ys_multiclass, [MULTICLASS_CLASSIFICATION]),
            itertools.product(ys_multilabel, [MULTILABEL_CLASSIFICATION]),
            itertools.product(ys_regression, [REGRESSION]),
            itertools.product(ys_multioutput_regression, [MULTIOUTPUT_REGRESSION]),
        )
    )


@pytest.mark.parametrize("X, y, X_test, y_test, task", data_input_and_target_types())
def test_input_and_target_types(dask_client, X, y, X_test, y_test, task):

    if task in CLASSIFICATION_TASKS:
        automl = AutoMLClassifier(
            time_left_for_this_task=15,
            per_run_time_limit=5,
            dask_client=dask_client,
        )
    else:
        automl = AutoMLRegressor(
            time_left_for_this_task=15,
            per_run_time_limit=5,
            dask_client=dask_client,
        )
    # To save time fitting and only validate the inputs we only return
    # the configuration space
    automl.fit(
        X=X,
        y=y,
        X_test=X_test,
        y_test=y_test,
        only_return_configuration_space=True
    )
    assert automl._task == task
    assert automl._metric.name == default_metric_for_task[task].name


def data_test_model_predict_outsputs_correct_shapes():
    datasets = sklearn.datasets
    binary = datasets.make_classification(
        n_samples=5, n_classes=2, random_state=0
    )
    multiclass = datasets.make_classification(
        n_samples=5, n_informative=3, n_classes=3, random_state=0
    )
    multilabel = datasets.make_multilabel_classification(
        n_samples=5, n_classes=3, random_state=0
    )
    regression = datasets.make_regression(
        n_samples=5, random_state=0
    )
    multioutput = datasets.make_regression(
        n_samples=5, n_targets=3, random_state=0
    )

    # TODO issue 1169
    #   While testing output shapes, realised all models are wrapped to provide
    #   a special predict_proba that outputs a different shape than usual.
    #   This includes DummyClassifier and DummyRegressor which are wrapped as
    #   `MyDummyClassifier/Regressor` and require a config object.
    #   config == 1 : Classifier uses 'uniform', Regressor uses 'mean'
    #   else        : Classifier uses 'most_frequent', Regressor uses 'median'
    #
    #   This wrapping of probabilities with
    #       `convert_multioutput_multiclass_to_multilabel`
    #   can probably be just put into a base class which queries subclasses
    #   as to whether it's needed.
    #
    #   tldr; thats why we use MyDummyX here instead of the default models
    #         from sklearn
    def classifier(X, y):
        return MyDummyClassifier(config=1, random_state=0).fit(X, y)

    def regressor(X, y):
        return MyDummyRegressor(config=1, random_state=0).fit(X, y)

    # How cross validation models are currently grouped together
    def voting_classifier(X, y):
        classifiers = [
            MyDummyClassifier(config=1, random_state=0).fit(X, y)
            for _ in range(5)
        ]
        vc = VotingClassifier(estimators=None, voting='soft')
        vc.estimators_ = classifiers
        return vc

    def voting_regressor(X, y):
        regressors = [
            MyDummyRegressor(config=1, random_state=0).fit(X, y)
            for _ in range(5)
        ]
        vr = VotingRegressor(estimators=None)
        vr.estimators_ = regressors
        return vr

    test_data = {
        BINARY_CLASSIFICATION: {
            'models': [classifier(*binary), voting_classifier(*binary)],
            'data': binary,
            # prob of false/true for the one class
            'expected_output_shape': (len(binary[0]), 2)
        },
        MULTICLASS_CLASSIFICATION: {
            'models': [classifier(*multiclass), voting_classifier(*multiclass)],
            'data': multiclass,
            # prob of true for each possible class
            'expected_output_shape': (len(multiclass[0]), 3)
        },
        MULTILABEL_CLASSIFICATION: {
            'models': [classifier(*multilabel), voting_classifier(*multilabel)],
            'data': multilabel,
            # probability of true for each binary label
            'expected_output_shape': (len(multilabel[0]), 3)  # type: ignore
        },
        REGRESSION: {
            'models': [regressor(*regression), voting_regressor(*regression)],
            'data': regression,
            # array of single outputs
            'expected_output_shape': (len(regression[0]), )
        },
        MULTIOUTPUT_REGRESSION: {
            'models': [regressor(*multioutput), voting_regressor(*multioutput)],
            'data': multioutput,
            # array of vector otuputs
            'expected_output_shape': (len(multioutput[0]), 3)
        }
    }

    return itertools.chain.from_iterable(
        [
            (model, cfg['data'], task, cfg['expected_output_shape'])
            for model in cfg['models']
        ]
        for task, cfg in test_data.items()
    )


@pytest.mark.parametrize(
    "model, data, task, expected_output_shape",
    data_test_model_predict_outsputs_correct_shapes()
)
def test_model_predict_outputs_correct_shapes(model, data, task, expected_output_shape):
    X, y = data
    prediction = _model_predict(model=model, X=X, task=task)
    assert prediction.shape == expected_output_shape


def test_model_predict_outputs_warnings_to_logs():
    X = list(range(20))
    task = REGRESSION
    logger = PickableLoggerAdapter('test_model_predict_correctly_outputs_warnings')
    logger.warning = unittest.mock.Mock()

    class DummyModel:
        def predict(self, x):
            warnings.warn('test warning', Warning)
            return x

    model = DummyModel()

    _model_predict(model, X, task, logger=logger)

    assert logger.warning.call_count == 1, "Logger should have had warning called"


def test_model_predict_outputs_to_stdout_if_no_logger():
    X = list(range(20))
    task = REGRESSION

    class DummyModel:
        def predict(self, x):
            warnings.warn('test warning', Warning)
            return x

    model = DummyModel()

    with warnings.catch_warnings(record=True) as w:
        _model_predict(model, X, task, logger=None)

        assert len(w) == 1, "One warning sould have been emmited"


@pytest.mark.parametrize("task, y", [
    (BINARY_CLASSIFICATION, np.asarray(
        9999 * [0] + 1 * [1]
    )),
    (MULTICLASS_CLASSIFICATION, np.asarray(
        4999 * [1] + 4999 * [2] + 1 * [3] + 1 * [4]
    )),
    (MULTILABEL_CLASSIFICATION, np.asarray(
        4999 * [[0, 1, 1]] + 4999 * [[1, 1, 0]] + 1 * [[1, 0, 1]] + 1 * [[0, 0, 0]]
    ))
])
def test_subsample_classification_unique_labels_stay_in_training_set(task, y):
    n_samples = 10000
    X = np.random.random(size=(n_samples, 3))
    memory_limit = 1  # Force subsampling
    mock = unittest.mock.Mock()

    # Make sure our test assumptions are correct
    assert len(y) == n_samples, "Ensure tests are correctly setup"

    values, counts = np.unique(y, axis=0, return_counts=True)
    unique_labels = [value for value, count in zip(values, counts) if count == 1]
    assert len(unique_labels), "Ensure we have unique labels in the test"

    _, y_sampled = AutoML.subsample_if_too_large(X, y,
                                                 logger=mock,
                                                 seed=1,
                                                 memory_limit=memory_limit,
                                                 task=task)

    assert len(y_sampled) <= len(y), \
        "Ensure sampling took place"
    assert all(label in y_sampled for label in unique_labels), \
        "All unique labels present in the return sampled set"



import numpy as np

import pytest

from sklearn.base import clone
from sklearn.datasets import make_classification, make_regression
from sklearn.utils.validation import check_is_fitted

from autosklearn.evaluation.abstract_evaluator import MyDummyClassifier, MyDummyRegressor


@pytest.mark.parametrize("task_type", ['classification', 'regression'])
def test_dummy_pipeline(task_type):
    if task_type == 'classification':
        estimator_class = MyDummyClassifier
        data_maker = make_classification
    elif task_type == 'regression':
        estimator_class = MyDummyRegressor
        data_maker = make_regression
    else:
        pytest.fail(task_type)
        return

    estimator = estimator_class(config=1, random_state=0)
    X, y = data_maker(random_state=0)
    estimator.fit(X, y)
    check_is_fitted(estimator)

    assert np.shape(X)[0] == np.shape(estimator.predict(X))[0]

    # make sure we comply with scikit-learn estimator API
    clone(estimator)



import logging
import numpy as np
import os
import unittest

import pandas as pd

from ConfigSpace.configuration_space import Configuration
import autosklearn.pipeline.classification

import autosklearn.metalearning.optimizers.metalearn_optimizer.metalearner as metalearner
from autosklearn.metalearning.metalearning.meta_base import MetaBase

logging.basicConfig()


class MetaLearnerTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def setUp(self):
        self.cwd = os.getcwd()
        data_dir = os.path.dirname(__file__)
        data_dir = os.path.join(data_dir, 'test_meta_base_data')
        os.chdir(data_dir)

        pipeline = autosklearn.pipeline.classification.SimpleClassificationPipeline()
        self.cs = pipeline.get_hyperparameter_search_space()

        self.logger = logging.getLogger()
        meta_base = MetaBase(self.cs, data_dir, logger=self.logger)
        self.meta_optimizer = metalearner.MetaLearningOptimizer(
            '233', self.cs, meta_base, logger=self.logger)

    def tearDown(self):
        os.chdir(self.cwd)

    def test_metalearning_suggest_all(self):
        ret = self.meta_optimizer.metalearning_suggest_all()
        self.assertEqual(124, len(ret))
        # Reduced to 17 as we changed QDA searchspace
        self.assertEqual('gradient_boosting', ret[0]['classifier:__choice__'])
        self.assertEqual('adaboost', ret[1]['classifier:__choice__'])
        # There is no test for exclude_double_configuration as it's not present
        # in the test data

    def test_metalearning_suggest_all_nan_metafeatures(self):
        self.meta_optimizer.meta_base.metafeatures.loc["233"].iloc[:10] = np.NaN
        ret = self.meta_optimizer.metalearning_suggest_all()
        self.assertEqual(124, len(ret))
        # Reduced to 17 as we changed QDA searchspace
        self.assertEqual('gradient_boosting', ret[0]['classifier:__choice__'])
        self.assertEqual('gradient_boosting', ret[1]['classifier:__choice__'])

    def test_metalearning_suggest(self):
        ret = self.meta_optimizer.metalearning_suggest([])
        self.assertIsInstance(ret, Configuration)
        self.assertEqual('gradient_boosting', ret['classifier:__choice__'])

        ret2 = self.meta_optimizer.metalearning_suggest([ret])
        self.assertIsInstance(ret2, Configuration)
        self.assertEqual('adaboost', ret2['classifier:__choice__'])

    def test_learn(self):
        # Test only some special cases which are probably not yet handled
        # like the metafeatures to eliminate and the random forest
        # hyperparameters
        self.meta_optimizer._learn()

    def test_split_metafeature_array(self):
        ds_metafeatures, other_metafeatures = self.meta_optimizer. \
            _split_metafeature_array()
        self.assertIsInstance(ds_metafeatures, pd.Series)
        self.assertEqual(ds_metafeatures.shape, (46,))
        self.assertIsInstance(other_metafeatures, pd.DataFrame)
        self.assertEqual(other_metafeatures.shape, (131, 46))


if __name__ == "__main__":
    unittest.main()



import sklearn.ensemble

from autosklearn.pipeline.components.regression.extra_trees import \
    ExtraTreesRegressor
from .test_base import BaseRegressionComponentTest


class ExtraTreesComponentTest(BaseRegressionComponentTest):

    __test__ = True

    res = dict()
    res["default_boston"] = 0.8539264243687228
    res["boston_n_calls"] = 9
    res["default_boston_iterative"] = res['default_boston']
    res["default_boston_sparse"] = 0.411211701806908
    res["default_boston_iterative_sparse"] = res['default_boston_sparse']
    res["default_diabetes"] = 0.3885150255877827
    res["diabetes_n_calls"] = 9
    res["default_diabetes_iterative"] = res['default_diabetes']
    res["default_diabetes_sparse"] = 0.2422804139169642
    res["default_diabetes_iterative_sparse"] = res['default_diabetes_sparse']

    sk_mod = sklearn.ensemble.ExtraTreesRegressor
    module = ExtraTreesRegressor
    step_hyperparameter = {
        'name': 'n_estimators',
        'value': module.get_max_iter(),
    }



__author__ = 'feurerm'

import copy
import unittest

import numpy as np
import sklearn.datasets
import sklearn.metrics

from autosklearn.pipeline.components.data_preprocessing.balancing.balancing \
    import Balancing
from autosklearn.pipeline.classification import SimpleClassificationPipeline
from autosklearn.pipeline.components.classification.adaboost import AdaboostClassifier
from autosklearn.pipeline.components.classification.decision_tree import DecisionTree
from autosklearn.pipeline.components.classification.extra_trees import ExtraTreesClassifier
from autosklearn.pipeline.components.classification.random_forest import RandomForest
from autosklearn.pipeline.components.classification.liblinear_svc import LibLinear_SVC
from autosklearn.pipeline.components.classification.libsvm_svc import LibSVM_SVC
from autosklearn.pipeline.components.classification.sgd import SGD
from autosklearn.pipeline.components.classification.gradient_boosting \
    import GradientBoostingClassifier
from autosklearn.pipeline.components.classification.passive_aggressive import PassiveAggressive
from autosklearn.pipeline.components.feature_preprocessing\
    .extra_trees_preproc_for_classification import ExtraTreesPreprocessorClassification
from autosklearn.pipeline.components.feature_preprocessing.liblinear_svc_preprocessor import \
    LibLinear_Preprocessor


class BalancingComponentTest(unittest.TestCase):
    def test_balancing_get_weights_treed_single_label(self):
        Y = np.array([0] * 80 + [1] * 20)
        balancing = Balancing(strategy='weighting')
        init_params, fit_params = balancing.get_weights(
            Y, 'adaboost', None, None, None)
        self.assertAlmostEqual(
            np.mean(fit_params['classifier:sample_weight']), 1,
        )
        np.testing.assert_allclose(
            fit_params['classifier:sample_weight'],
            np.array([0.625] * 80 + [2.5] * 20),
        )

    def test_balancing_get_weights_treed_multilabel(self):
        Y = np.array([[0, 0, 0]] * 100 + [[1, 0, 0]] * 100 + [[0, 1, 0]] * 100 +
                     [[1, 1, 0]] * 100 + [[0, 0, 1]] * 100 + [[1, 0, 1]] * 10)
        balancing = Balancing(strategy='weighting')
        init_params, fit_params = balancing.get_weights(
            Y, 'adaboost', None, None, None)
        print(fit_params['classifier:sample_weight'])
        self.assertAlmostEqual(
            np.mean(fit_params['classifier:sample_weight']), 1,
        )
        np.testing.assert_allclose(
            fit_params['classifier:sample_weight'],
            np.array([0.85] * 500 + [8.5] * 10),
        )

    def test_balancing_get_weights_svm_sgd(self):
        Y = np.array([0] * 80 + [1] * 20)
        balancing = Balancing(strategy='weighting')
        init_params, fit_params = balancing.get_weights(
            Y, 'libsvm_svc', None, None, None)
        self.assertEqual(("classifier:class_weight", "balanced"),
                         list(init_params.items())[0])
        init_params, fit_params = balancing.get_weights(
            Y, None, 'liblinear_svc_preprocessor', None, None)
        self.assertEqual(("feature_preprocessor:class_weight", "balanced"),
                         list(init_params.items())[0])

    def test_weighting_effect(self):
        data = sklearn.datasets.make_classification(
            n_samples=200, n_features=10, n_redundant=2, n_informative=2,
            n_repeated=2, n_clusters_per_class=2, weights=[0.8, 0.2],
            random_state=1)

        for name, clf, acc_no_weighting, acc_weighting, places in [
            ('adaboost', AdaboostClassifier, 0.810, 0.735, 3),
            ('decision_tree', DecisionTree, 0.780, 0.643, 3),
            ('extra_trees', ExtraTreesClassifier, 0.78, 0.8, 3),
            ('random_forest', RandomForest, 0.75, 0.789, 3),
            ('libsvm_svc', LibSVM_SVC, 0.769, 0.72, 3),
            ('liblinear_svc', LibLinear_SVC, 0.762, 0.735, 3),
            ('passive_aggressive', PassiveAggressive, 0.16, 0.222, 3),
            ('sgd', SGD, 0.818, 0.567, 2),
            ('gradient_boosting', GradientBoostingClassifier, 0.666, 0.682, 2)
         ]:
            for strategy, acc in [
                ('none', acc_no_weighting),
                ('weighting', acc_weighting)
            ]:
                # Fit
                data_ = copy.copy(data)
                X_train = data_[0][:100]
                Y_train = data_[1][:100]
                X_test = data_[0][100:]
                Y_test = data_[1][100:]

                model_args = {
                    "random_state": 1,
                    "include": {
                        'classifier': [name],
                        'feature_preprocessor': ['no_preprocessing']
                    }
                }

                classifier = SimpleClassificationPipeline(**model_args)
                cs = classifier.get_hyperparameter_search_space()
                default = cs.get_default_configuration()
                default._values['balancing:strategy'] = strategy

                classifier = SimpleClassificationPipeline(config=default, **model_args)
                classifier.fit(X_train, Y_train)

                predictions1 = classifier.predict(X_test)
                self.assertAlmostEqual(
                    sklearn.metrics.f1_score(predictions1, Y_test), acc,
                    places=places, msg=(name, strategy)
                )

                # fit_transformer and fit_estimator
                data_ = copy.copy(data)
                X_train = data_[0][:100]
                Y_train = data_[1][:100]
                X_test = data_[0][100:]
                Y_test = data_[1][100:]

                classifier = SimpleClassificationPipeline(config=default, **model_args)
                Xt, fit_params = classifier.fit_transformer(X_train, Y_train)
                classifier.fit_estimator(Xt, Y_train, **fit_params)

                predictions2 = classifier.predict(X_test)
                np.testing.assert_allclose(
                    predictions1, predictions2,
                    err_msg=f"name = {name}, strategy = {strategy}"
                )
                self.assertAlmostEqual(
                    sklearn.metrics.f1_score(predictions2, Y_test), acc,
                    places=places, msg=(name, strategy)
                )

        for name, pre, acc_no_weighting, acc_weighting in \
                [('extra_trees_preproc_for_classification',
                    ExtraTreesPreprocessorClassification, 0.810, 0.590),
                 ('liblinear_svc_preprocessor', LibLinear_Preprocessor,
                    0.837, 0.562)]:
            for strategy, acc in [('none', acc_no_weighting),
                                  ('weighting', acc_weighting)]:
                data_ = copy.copy(data)
                X_train = data_[0][:100]
                Y_train = data_[1][:100]
                X_test = data_[0][100:]
                Y_test = data_[1][100:]

                include = {'classifier': ['sgd'], 'feature_preprocessor': [name]}

                classifier = SimpleClassificationPipeline(random_state=1, include=include)
                cs = classifier.get_hyperparameter_search_space()
                default = cs.get_default_configuration()
                default._values['balancing:strategy'] = strategy
                classifier.set_hyperparameters(default)
                predictor = classifier.fit(X_train, Y_train)
                predictions = predictor.predict(X_test)
                self.assertAlmostEqual(
                    sklearn.metrics.f1_score(predictions, Y_test), acc,
                    places=3, msg=(name, strategy))

                # fit_transformer and fit_estimator
                data_ = copy.copy(data)
                X_train = data_[0][:100]
                Y_train = data_[1][:100]
                X_test = data_[0][100:]
                Y_test = data_[1][100:]

                default._values['balancing:strategy'] = strategy
                classifier = SimpleClassificationPipeline(default, random_state=1, include=include)
                Xt, fit_params = classifier.fit_transformer(X_train, Y_train)
                classifier.fit_estimator(Xt, Y_train, **fit_params)
                predictions = classifier.predict(X_test)
                self.assertAlmostEqual(
                    sklearn.metrics.f1_score(predictions, Y_test), acc,
                    places=3)



# -*- encoding: utf-8 -*-
import re
import os
import glob
import typing

import numpy as np


scores_dict = {
    'train_single': ["single_best_train_score", "single_best_optimization_score"],
    'test_single': ["single_best_test_score"],
    'train_ensamble': ["ensemble_optimization_score"],
    'test_ensamble': ["ensemble_test_score"]
}


def print_debug_information(automl):

    # In case it is called with estimator,
    # Get the automl object
    if hasattr(automl, 'automl_'):
        automl = automl.automl_

    # Log file path
    log_file = glob.glob(os.path.join(
        automl._backend.temporary_directory, 'AutoML*.log'))[0]

    include_messages = ['INFO', 'DEBUG', 'WARN',
                        'CRITICAL', 'ERROR', 'FATAL']

    # There is a lot of content in the log files. Only
    # parsing the main message and ignore the metalearning
    # messages
    try:
        with open(log_file) as logfile:
            content = logfile.readlines()

        # Get the messages to debug easier!
        content = [line for line in content if any(
            msg in line for msg in include_messages
        ) and 'metalearning' not in line]

    except Exception as e:
        return str(e)

    # Also add the run history if any
    if hasattr(automl, 'runhistory_') and hasattr(automl.runhistory_, 'data'):
        for k, v in automl.runhistory_.data.items():
            content += ["{}->{}".format(k, v)]
    else:
        content += ['No RunHistory']

    # Also add the ensemble history if any
    if len(automl.ensemble_performance_history) > 0:
        content += [str(h) for h in automl.ensemble_performance_history]
    else:
        content += ['No Ensemble History']

    return os.linesep.join(content)


def _includes(scores, all_scores):
    return all(score in all_scores for score in scores) and len(scores) == len(all_scores)


def count_succeses(cv_results):
    return np.sum(
        [status in ['Success', 'Success (but do not advance to higher budget)']
         for status in cv_results['status']]
    )


def includes_all_scores(scores):
    all_scores = scores_dict["train_single"] + scores_dict["test_single"] + \
                 scores_dict["train_ensamble"] + scores_dict["test_ensamble"] + ["Timestamp"]
    return _includes(scores, all_scores)


def include_single_scores(scores):
    all_scores = scores_dict["train_single"] + scores_dict["test_single"] + ["Timestamp"]
    return _includes(scores, all_scores)


def includes_train_scores(scores):
    all_scores = scores_dict["train_single"] + scores_dict["train_ensamble"] + ["Timestamp"]
    return _includes(scores, all_scores)


def performance_over_time_is_plausible(poT):
    if len(poT) < 1:
        return False
    if len(poT.drop(columns=["Timestamp"]).dropna()) < 1:
        return False
    if not poT["Timestamp"].is_monotonic:
        return False
    return True


class AutoMLLogParser(object):
    def __init__(self, logfile: str):
        self.logfile = logfile
        self.lines = self.parse_logfile()

    def parse_logfile(self) -> typing.List[str]:
        # We care about the [debug/info/...] messages
        # At the time of writing, the message format was:
        # [DEBUG] [2020-11-30 11:54:05,072:EnsembleBuilder] Restricting your
        # function to 3072 mb memory.
        #
        # [DEBUG] [2020-11-30 11:53:55,062:pynisher] Redirecting
        # output of the function to files.
        assert os.path.exists(self.logfile), "{} not found".format(self.logfile)

        with open(self.logfile) as fh:
            content = [line.strip() for line in fh if re.search(r'[\w+]', line)]
        return content

    def count_ensembler_iterations(self) -> int:
        iterations = []

        # One thing is to call phynisher, the other is to actually execute the funciton
        iterations_from_inside_ensemble_builder = []
        for line in self.lines:

            # Pynisher call
            # we have to count the start msg from pynisher
            # and the return msg
            # We expect the start msg to be something like:
            # [DEBUG] [2020-11-26 19:22:42,160:EnsembleBuilder] \
            # Function called with argument: (61....
            # [DEBUG] [2020-11-30 11:53:47,069:EnsembleBuilder] Function called with argument:
            # (28.246965646743774, 1, False), {}
            match = re.search(
                r'EnsembleBuilder]\s+Function called with argument:\s+\(\d+\.\d+, (\d+), \w+',
                line)
            if match:
                iterations.append(int(match.group(1)))

            # Ensemble Builder actual call
            # Here we expect the msg:
            # [DEBUG] [2020-11-30 11:53:14,877:EnsembleBuilder] Starting iteration 0,
            # time left: 61.266255
            # [DEBUG] [2020-11-27 20:27:28,044:EnsembleBuilder] Starting iteration 2,
            # time left: 10.603252
            match = re.search(
                r'EnsembleBuilder]\s+Starting iteration (\d+)',
                line)
            if match:
                iterations_from_inside_ensemble_builder.append(int(match.group(1)))

            # The ensemble builder might not be called if there is no time.
            # Here we expect the msg:
            # [DEBUG] [2020-11-27 20:27:28,044:EnsembleBuilder] Not starting iteration 2,
            # as time left: 1.59324
            match = re.search(
                r'EnsembleBuilder]\s+Not starting iteration (\d+)',
                line)
            if match:
                iterations_from_inside_ensemble_builder.append(int(match.group(1)))

        assert iterations == iterations_from_inside_ensemble_builder, "{} ! {}".format(
            iterations, iterations_from_inside_ensemble_builder
        )

        return iterations

    def count_ensembler_success_pynisher_calls(self) -> int:

        # We expect the return msg to be something like:
        # [DEBUG] [2020-11-30 11:53:47,911:EnsembleBuilder] return value:
        # (([{'Timestamp': Timestamp('2020-11-30 11:53:47.910727'),
        # 'ensemble_optimization_score': 0.9787234042553191}], 50, None, None, None), 0)
        # [DEBUG] [2020-11-30 11:54:05,984:EnsembleBuilder] return value:
        # (([{'Timestamp': Timestamp('2020-11- 30 11:54:05.983837'),
        # 'ensemble_optimization_score': 0.9787234042553191}], 50, None, None, None), 0)
        return_msgs = len([line for line in self.lines if re.search(
            r'EnsembleBuilder]\s+return value:.*Timestamp', line)])

        return return_msgs

    def count_tae_pynisher_calls(self) -> int:
        # We expect the return msg to be something like:
        # [DEBUG] [2020-12-16 11:57:08,987:Client-pynisher] Function called with argument: ()
        # , {'queue': <multiprocessing.queues.Queue object at 0x7f9e3cfaae20>, 'config': 1
        # [DEBUG] [2020-12-16 11:57:10,537:Client-pynisher] Function called with argument: ()
        # , {'queue': <multiprocessing.queues.Queue object at 0x7f16f5d95c40>,
        # 'config': Configuration:
        # Only the parenthesis below need to be escaped, ] and { do not.
        call_msgs = len([line for line in self.lines if re.search(
            r'pynisher]\s+Function called with argument: \(\), {', line)])
        return call_msgs

    def count_tae_pynisher_returns(self) -> int:
        # We expect the return msg to be something like:
        # [DEBUG] [2020-11-30 11:53:11,264:pynisher] return value: (None, 0)
        # [DEBUG] [2020-11-30 11:53:13,768:pynisher] return value: (None, 0)
        return_msgs = len([line for line in self.lines if re.search(
            r'pynisher]\s+return value:\s+', line)])
        # When the pynisher pipe is prematurely closed, we also expect:
        # Your function call closed the pipe prematurely
        # -> Subprocess probably got an uncatchable signal
        # We expect the return msg to be something like:
        # OR
        # Something else went wrong, sorry.
        premature_msgs = len([line for line in self.lines if re.search(
            r'pynisher]\s+Your function call closed the pipe prematurely', line)])
        failure_msgs = len([line for line in self.lines if re.search(
            r'pynisher]\s+Something else went wrong, sorry.', line)])
        return return_msgs + premature_msgs + failure_msgs

    def get_automl_setting_from_log(self, dataset_name: str, setting: str) -> str:
        for line in self.lines:
            # We expect messages of the form
            # [DEBUG] [2020-11-30 11:53:10,457:AutoML(5):breast_cancer]   ensemble_size: 50
            # [DEBUG] [2020-11-30 11:53:10,457:AutoML(5):breast_cancer]   ensemble_nbest: 50
            match = re.search(
                f"{dataset_name}]\\s*{setting}\\s*:\\s*(\\w+)",
                line)
            if match:
                return match.group(1)
        return None



import json
import os
import random
import shutil
import socket
import subprocess
import unittest

import arff
import numpy as np

from autosklearn.metrics import CLASSIFICATION_METRICS, REGRESSION_METRICS


class TestMetadataGeneration(unittest.TestCase):

    def setUp(self):
        self.working_directory = '/tmp/autosklearn-unittest-tmp-dir-%s-%d-%d' % (
            socket.gethostname(), os.getpid(), random.randint(0, 1000000))

    def print_files(self):
        print('Existing files:')
        for dirpath, dirnames, filenames in os.walk(self.working_directory):
            print(dirpath, dirnames, filenames)

    def test_metadata_generation(self):

        regression_task_id = 360029
        regression_dataset_name = 'SWD'.lower()
        classification_task_id = 245
        classification_dataset_name = 'breast-w'.lower()

        current_directory = __file__
        scripts_directory = os.path.abspath(os.path.join(current_directory,
                                                         '..', '..', '..',
                                                         'scripts'))

        # 1. create working directory
        try:
            os.makedirs(self.working_directory)
        except Exception as e:
            print(e)

        # 2. should be done by the person running the unit tests!

        # 3. create configuration commands
        script_filename = os.path.join(scripts_directory, '01_create_commands.py')
        cmd = 'python3 %s --working-directory %s --test' % (script_filename, self.working_directory)
        rval = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.assertEqual(rval.returncode, 0, msg=str(rval))

        # 4. run one of the commands to get some data
        commands_output_file = os.path.join(self.working_directory, 'metadata_commands.txt')
        self.assertTrue(os.path.exists(commands_output_file))

        with open(commands_output_file) as fh:
            cmds = fh.read().split('\n')
            # 6 regression, 7 classification (roc_auc + task 258 is illegal), 1 empty line
            self.assertEqual(len(cmds), 18, msg='\n'.join(cmds))

        for task_id, dataset_name, task_type, metric in (
            (
                classification_task_id,
                classification_dataset_name,
                'classification',
                'balanced_accuracy',
            ),
            (regression_task_id, regression_dataset_name, 'regression', 'r2')
        ):
            cmd = None
            with open(commands_output_file) as fh:
                while True:
                    cmd = fh.readline()
                    if 'task-id %d' % task_id in cmd and metric in cmd:
                        break
            if cmd is None:
                self.fail('Did not find a command for task_id %s and metric %s in %s'
                          % (task_id, metric, cmds))

            self.assertIn('time-limit 86400', cmd)
            self.assertIn('per-run-time-limit 1800', cmd)
            cmd = cmd.replace('time-limit 86400', 'time-limit 60').replace(
                'per-run-time-limit 1800', 'per-run-time-limit 5')
            # This tells the script to use the same memory limit for testing as
            # for training. In production, it would use twice as much!
            cmd = cmd.replace('-s 1', '-s 1 --unittest')
            print('COMMAND: %s' % cmd)
            rval = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE)
            print('STDOUT: %s' % repr(rval.stdout), flush=True)
            print('STDERR: %s' % repr(rval.stderr), flush=True)

            self.print_files()

            expected_output_directory = os.path.join(self.working_directory,
                                                     'configuration',
                                                     task_type,
                                                     str(task_id), metric,
                                                     'auto-sklearn-output')
            self.assertTrue(os.path.exists(expected_output_directory),
                            msg=expected_output_directory)
            smac_log = os.path.join(expected_output_directory, 'AutoML(1):%s.log' % dataset_name)
            with open(smac_log) as fh:
                smac_output = fh.read()
            self.assertEqual(rval.returncode, 0, msg=str(rval) + '\n' + smac_output)
            expected_validation_output = os.path.join(expected_output_directory, '..',
                                                      'validation_trajectory_1.json')
            self.assertTrue(os.path.exists(expected_validation_output))
            trajectory = os.path.join(expected_output_directory,
                                      'smac3-output', 'run_1', 'trajectory.json')

            with open(expected_validation_output) as fh_validation:
                with open(trajectory) as fh_trajectory:
                    traj = json.load(fh_trajectory)
                    valid_traj = json.load(fh_validation)
                    print('Validation trajectory:')
                    print(valid_traj)
                    self.assertGreater(len(traj), 2, msg=str(valid_traj))
                    self.assertEqual(len(traj), len(valid_traj), msg=str(valid_traj))
                    for entry in valid_traj:
                        if task_type == 'classification':
                            for metric in CLASSIFICATION_METRICS:
                                # This is a multilabel metric
                                if metric in ('precision_samples', 'recall_samples', 'f1_samples'):
                                    continue
                                self.assertIn(metric, entry[-1])
                                self.assertIsInstance(entry[-1][metric], float)
                                self.assertTrue(np.isfinite(entry[-1][metric]),
                                                (metric, str(entry[-1][metric])))
                        else:
                            for metric in REGRESSION_METRICS:
                                self.assertIn(metric, entry[-1])
                                self.assertIsInstance(entry[-1][metric], float)
                                self.assertTrue(np.isfinite(entry[-1][metric]),
                                                (metric, str(entry[-1][metric])))

        # 5. Get the test performance of these configurations
        script_filename = os.path.join(scripts_directory, '02_retrieve_metadata.py')
        cmd = 'python3 %s --working-directory %s ' % (script_filename, self.working_directory)
        print('COMMAND: %s' % cmd)
        rval = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        print('STDOUT: %s' % repr(rval.stdout), flush=True)
        print('STDERR: %s' % repr(rval.stderr), flush=True)
        self.assertEqual(rval.returncode, 0, msg=str(rval))

        for file in ['algorithm_runs.arff', 'configurations.csv', 'description.results.txt']:
            for metric in ['accuracy', 'balanced_accuracy', 'log_loss', 'roc_auc']:
                path = os.path.join(
                    self.working_directory,
                    'configuration_results',
                    '%s_binary.classification_dense' % metric,
                    file,
                )
                self.assertTrue(os.path.exists(path), msg=path)

        for file in ['algorithm_runs.arff', 'configurations.csv', 'description.results.txt']:
            for metric in ['r2', 'mean_squared_error']:
                path = os.path.join(
                    self.working_directory,
                    'configuration_results',
                    '%s_regression_dense' % metric,
                    file,
                )
                self.assertTrue(os.path.exists(path), msg=path)

        # 6. Calculate metafeatures
        script_filename = os.path.join(scripts_directory, '03_calculate_metafeatures.py')
        cmd = (
            'python3 %s --working-directory %s --test-mode '
            % (script_filename, self.working_directory)
        )
        rval = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        self.assertEqual(rval.returncode, 0, msg=str(rval))
        for task_type in ('classification', 'regression'):
            for file in ['calculation_times.csv', 'description.features.txt',
                         'feature_costs.arff', 'feature_runstatus.arff',
                         'feature_values.arff']:
                self.assertTrue(
                    os.path.exists(os.path.join(
                        self.working_directory,
                        'metafeatures',
                        task_type,
                        file)
                    )
                )

        with open(
            os.path.join(
                self.working_directory, 'metafeatures', 'regression', 'feature_values.arff'
            )
        ) as fh:
            metafeatures_arff = fh.read().split('\n')
            contains_regression_id = False
            for line in metafeatures_arff:
                if line.startswith('fri_c4_500_25,'):
                    contains_regression_id = True
            self.assertTrue(contains_regression_id, msg=metafeatures_arff)

        with open(
                os.path.join(
                    self.working_directory, 'metafeatures', 'classification', 'feature_values.arff'
                )
        ) as fh:
            metafeatures_arff = fh.read().split('\n')
            contains_classification_id = False
            for line in metafeatures_arff:
                if line.startswith('anneal,'):
                    contains_classification_id = True
            self.assertTrue(contains_classification_id, msg=metafeatures_arff)

        # 7. Create aslib files
        script_filename = os.path.join(scripts_directory, '04_create_aslib_files.py')
        cmd = 'python3 %s --working-directory %s ' % (
            script_filename, self.working_directory)
        rval = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)
        self.assertEqual(rval.returncode, 0, msg=str(rval))

        for metric_, combination in (
            (metric, '%s_binary.classification_dense' % metric),
            (metric, '%s_regression_dense' % metric),
        ):

            if task_type not in combination:
                continue

            for file in ['algorithm_runs.arff', 'configurations.csv',
                         'description.txt', 'feature_costs.arff',
                         'feature_runstatus.arff', 'feature_values.arff',
                         'readme.txt']:
                expected_path = os.path.join(
                    self.working_directory, 'metadata', combination, file,
                )
                self.assertTrue(os.path.exists(expected_path), msg=expected_path)

            with open(os.path.join(self.working_directory,
                                   'metadata',
                                   combination,
                                   'algorithm_runs.arff')) as fh:
                algorithm_runs = arff.load(fh)
                self.assertEqual(algorithm_runs['attributes'],
                                 [('instance_id', 'STRING'),
                                  ('repetition', 'NUMERIC'),
                                  ('algorithm', 'STRING'),
                                  (metric_, 'NUMERIC'),
                                  ('runstatus',
                                   ['ok', 'timeout', 'memout', 'not_applicable',
                                    'crash', 'other'])])
                self.assertEqual(len(algorithm_runs['data']), 1)
                self.assertEqual(len(algorithm_runs['data'][0]), 5)
                self.assertLess(algorithm_runs['data'][0][3], 0.9)
                self.assertEqual(algorithm_runs['data'][0][4], 'ok')

    def tearDown(self):
        for i in range(5):
            try:
                shutil.rmtree(self.working_directory)
            except Exception:
                pass



import sklearn.metrics
import sklearn.svm

from autosklearn.pipeline.components.classification.libsvm_svc import LibSVM_SVC
from autosklearn.pipeline.util import get_dataset, \
    _test_classifier_predict_proba

from .test_base import BaseClassificationComponentTest


class LibSVM_SVCComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.96
    res["default_iris_iterative"] = -1
    res["default_iris_proba"] = 0.36298338197586716
    res["default_iris_sparse"] = 0.64
    res["default_digits"] = 0.096539162112932606
    res["default_digits_iterative"] = -1
    res["default_digits_binary"] = 0.90103217972070426
    res["default_digits_multilabel"] = -1
    res["default_digits_multilabel_proba"] = -1

    sk_mod = sklearn.svm.SVC
    module = LibSVM_SVC

    def test_default_configuration_predict_proba_individual(self):
        # Leave this additional test here
        for i in range(2):
            predictions, targets = _test_classifier_predict_proba(
                LibSVM_SVC, sparse=True, dataset='digits',
                train_size_maximum=500)
            self.assertAlmostEqual(5.273502056835706,
                                   sklearn.metrics.log_loss(targets,
                                                            predictions))

        for i in range(2):
            predictions, targets = _test_classifier_predict_proba(
                LibSVM_SVC, sparse=True, dataset='iris')
            self.assertAlmostEqual(0.8408320837510618,
                                   sklearn.metrics.log_loss(targets,
                                                            predictions))

        # 2 class
        for i in range(2):
            X_train, Y_train, X_test, Y_test = get_dataset(dataset='iris')
            remove_training_data = Y_train == 2
            remove_test_data = Y_test == 2
            X_train = X_train[~remove_training_data]
            Y_train = Y_train[~remove_training_data]
            X_test = X_test[~remove_test_data]
            Y_test = Y_test[~remove_test_data]
            ss = sklearn.preprocessing.StandardScaler()
            X_train = ss.fit_transform(X_train)
            configuration_space = LibSVM_SVC.get_hyperparameter_search_space()
            default = configuration_space.get_default_configuration()

            cls = LibSVM_SVC(random_state=1, **{hp_name: default[hp_name]
                                                for hp_name in default
                                                if default[hp_name] is not None})

            cls = cls.fit(X_train, Y_train)
            prediction = cls.predict_proba(X_test)
            self.assertAlmostEqual(sklearn.metrics.log_loss(Y_test, prediction),
                                   0.6927962762794081, places=4)



import unittest

import autosklearn.pipeline.components.feature_preprocessing as fp


class FeatureProcessingTest(unittest.TestCase):
    def test_get_available_components(self):
        # Target type
        for target_type, num_values in [('classification', 15),
                                        ('regression', 14)]:
            data_properties = {'target_type': target_type}

            available_components = fp.FeaturePreprocessorChoice(data_properties)\
                .get_available_components(data_properties)

            self.assertEqual(len(available_components), num_values)

        # Multiclass
        data_properties = {'target_type': 'classification',
                           'multiclass': True}
        available_components = fp.FeaturePreprocessorChoice(data_properties) \
            .get_available_components(data_properties)

        self.assertEqual(len(available_components), 15)

        # Multilabel
        data_properties = {'target_type': 'classification',
                           'multilabel': True}
        available_components = fp.FeaturePreprocessorChoice(data_properties) \
            .get_available_components(data_properties)

        self.assertEqual(len(available_components), 12)



import functools
import traceback
import tempfile
import unittest

import numpy as np
from numpy.linalg import LinAlgError
import sklearn.datasets
from sklearn import preprocessing
import sklearn.model_selection

from autosklearn.automl_common.common.utils.backend import Backend

from autosklearn.constants import \
    MULTICLASS_CLASSIFICATION, MULTILABEL_CLASSIFICATION, BINARY_CLASSIFICATION, REGRESSION
from autosklearn.util.data import convert_to_bin
from autosklearn.data.xy_data_manager import XYDataManager
from autosklearn.pipeline.util import get_dataset
from autosklearn.metrics import accuracy, balanced_accuracy, f1_macro, f1_micro, f1_weighted, \
    log_loss, precision_macro, precision_micro, precision_weighted, recall_macro, \
    recall_micro, recall_weighted

SCORER_LIST = [accuracy, balanced_accuracy, f1_macro, f1_micro, f1_weighted, log_loss,
               precision_macro, precision_micro, precision_weighted, recall_macro,
               recall_micro, recall_weighted]

N_TEST_RUNS = 5


def get_evaluation_backend():
    backend_mock = unittest.mock.Mock(spec=Backend)
    backend_mock.temporary_directory = tempfile.gettempdir()

    # Assign a default data
    backend_mock.load_datamanager.return_value = get_multiclass_classification_datamanager()

    return backend_mock


class Dummy(object):
    def __init__(self):
        self.name = 'Dummy'


class BaseEvaluatorTest(unittest.TestCase):
    def __init__(self, methodName):
        super(BaseEvaluatorTest, self).__init__(methodName)
        self.output_directories = []

    def _fit(self, evaluator):
        return self.__fit(evaluator.fit)

    def _partial_fit(self, evaluator, fold):
        partial_fit = functools.partial(evaluator.partial_fit, fold=fold)
        return self.__fit(partial_fit)

    def __fit(self, function_handle):
        """Allow us to catch known and valid exceptions for all evaluate
        scripts."""
        try:
            function_handle()
            return True
        except KeyError as e:
            if 'Floating-point under-/overflow occurred at epoch' in \
                    e.args[0] or \
                    'removed all features' in e.args[0] or \
                    'failed to create intent' in e.args[0]:
                pass
            else:
                traceback.print_exc()
                raise e
        except ValueError as e:
            if 'Floating-point under-/overflow occurred at epoch' in e.args[
                0] or \
                            'removed all features' in e.args[0] or \
                            'failed to create intent' in e.args[0]:
                pass
            else:
                raise e
        except LinAlgError as e:
            if 'not positive definite, even with jitter' in e.args[0]:
                pass
            else:
                raise e
        except RuntimeWarning as e:
            if 'invalid value encountered in sqrt' in e.args[0]:
                pass
            elif 'divide by zero encountered in divide' in e.args[0]:
                pass
            else:
                raise e
        except UserWarning as e:
            if 'FastICA did not converge' in e.args[0]:
                pass
            else:
                raise e


def get_multiclass_classification_datamanager():
    X_train, Y_train, X_test, Y_test = get_dataset('iris')
    indices = list(range(X_train.shape[0]))
    np.random.seed(1)
    np.random.shuffle(indices)
    X_train = X_train[indices]
    Y_train = Y_train[indices]

    X_valid = X_test[:25, ]
    Y_valid = Y_test[:25, ]
    X_test = X_test[25:, ]
    Y_test = Y_test[25:, ]

    D = Dummy()
    D.info = {
        'task': MULTICLASS_CLASSIFICATION,
        'is_sparse': False,
        'label_num': 3
    }
    D.data = {
        'X_train': X_train,
        'Y_train': Y_train,
        'X_valid': X_valid,
        'Y_valid': Y_valid,
        'X_test': X_test,
        'Y_test': Y_test
    }
    D.feat_type = {0: 'numerical',
                   1: 'Numerical',
                   2: 'numerical',
                   3: 'numerical'}
    return D


def get_abalone_datamanager():
    # https://www.openml.org/d/183
    dataset_name = 'abalone'
    data = sklearn.datasets.fetch_openml(data_id=183, as_frame=True)
    feat_type = {
        i: 'Categorical' if x.name == 'category' else 'Numerical'
        for i, x in enumerate(data['data'].dtypes)
    }
    X, y = sklearn.datasets.fetch_openml(data_id=183, return_X_y=True, as_frame=False)
    y = preprocessing.LabelEncoder().fit_transform(y)
    X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(
        X, y, random_state=1
    )

    D = XYDataManager(
        X_train, y_train,
        X_test, y_test,
        MULTICLASS_CLASSIFICATION,
        feat_type,
        dataset_name
    )
    return D


def get_multilabel_classification_datamanager():
    X_train, Y_train, X_test, Y_test = get_dataset('iris')
    indices = list(range(X_train.shape[0]))
    np.random.seed(1)
    np.random.shuffle(indices)
    X_train = X_train[indices]
    Y_train = Y_train[indices]

    Y_train = np.array(convert_to_bin(Y_train, 3))
    # for i in range(Y_train_.shape[0]):
    #    Y_train_[:, Y_train[i]] = 1
    # Y_train = Y_train_
    Y_test = np.array(convert_to_bin(Y_test, 3))
    # for i in range(Y_test_.shape[0]):
    #    Y_test_[:, Y_test[i]] = 1
    # Y_test = Y_test_

    X_valid = X_test[:25, ]
    Y_valid = Y_test[:25, ]
    X_test = X_test[25:, ]
    Y_test = Y_test[25:, ]

    D = Dummy()
    D.info = {
        'task': MULTILABEL_CLASSIFICATION,
        'is_sparse': False,
        'label_num': 3
    }
    D.data = {
        'X_train': X_train,
        'Y_train': Y_train,
        'X_valid': X_valid,
        'Y_valid': Y_valid,
        'X_test': X_test,
        'Y_test': Y_test
    }
    D.feat_type = {0: 'numerical',
                   1: 'Numerical',
                   2: 'numerical',
                   3: 'numerical'}
    return D


def get_binary_classification_datamanager():
    X_train, Y_train, X_test, Y_test = get_dataset('iris')
    indices = list(range(X_train.shape[0]))
    np.random.seed(1)
    np.random.shuffle(indices)
    X_train = X_train[indices]
    Y_train = Y_train[indices]

    eliminate_class_two = Y_train != 2
    X_train = X_train[eliminate_class_two]
    Y_train = Y_train[eliminate_class_two]

    eliminate_class_two = Y_test != 2
    X_test = X_test[eliminate_class_two]
    Y_test = Y_test[eliminate_class_two]

    X_valid = X_test[:25, ]
    Y_valid = Y_test[:25, ]
    X_test = X_test[25:, ]
    Y_test = Y_test[25:, ]

    D = Dummy()
    D.info = {
        'task': BINARY_CLASSIFICATION,
        'is_sparse': False,
        'label_num': 2
    }
    D.data = {
        'X_train': X_train,
        'Y_train': Y_train.reshape((-1, 1)),
        'X_valid': X_valid,
        'Y_valid': Y_valid.reshape((-1, 1)),
        'X_test': X_test,
        'Y_test': Y_test.reshape((-1, 1))
    }
    D.feat_type = {0: 'numerical',
                   1: 'Numerical',
                   2: 'numerical',
                   3: 'numerical'}
    return D


def get_regression_datamanager():
    X_train, Y_train, X_test, Y_test = get_dataset('boston')
    indices = list(range(X_train.shape[0]))
    np.random.seed(1)
    np.random.shuffle(indices)
    X_train = X_train[indices]
    Y_train = Y_train[indices]

    X_valid = X_test[:200, ]
    Y_valid = Y_test[:200, ]
    X_test = X_test[200:, ]
    Y_test = Y_test[200:, ]

    D = Dummy()
    D.info = {
        'task': REGRESSION,
        'is_sparse': False,
        'label_num': 1
    }
    D.data = {
        'X_train': X_train,
        'Y_train': Y_train.reshape((-1, 1)),
        'X_valid': X_valid,
        'Y_valid': Y_valid.reshape((-1, 1)),
        'X_test': X_test,
        'Y_test': Y_test.reshape((-1, 1))
    }
    D.feat_type = {i: 'numerical' for i in range(X_train.shape[1])}
    return D


def get_500_classes_datamanager():
    weights = ([0.002] * 475) + ([0.001] * 25)
    X, Y = sklearn.datasets.make_classification(n_samples=1000,
                                                n_features=20,
                                                n_classes=500,
                                                n_clusters_per_class=1,
                                                n_informative=15,
                                                n_redundant=5,
                                                n_repeated=0,
                                                weights=weights,
                                                flip_y=0,
                                                class_sep=1.0,
                                                hypercube=True,
                                                shift=None,
                                                scale=1.0,
                                                shuffle=True,
                                                random_state=1)

    D = Dummy()
    D.info = {
        'task': MULTICLASS_CLASSIFICATION,
        'is_sparse': False,
        'label_num': 500
    }
    D.data = {'X_train': X[:700], 'Y_train': Y[:700],
              'X_valid': X[700:710], 'Y_valid': Y[700:710],
              'X_test': X[710:], 'Y_test': Y[710:]
              }
    D.feat_type = {i: 'numerical' for i in range(20)}
    return D


def get_dataset_getters():
    return [get_binary_classification_datamanager,
            get_multiclass_classification_datamanager,
            get_multilabel_classification_datamanager,
            get_500_classes_datamanager,
            get_abalone_datamanager,
            get_regression_datamanager]



from typing import Optional, Dict

import unittest

from autosklearn.pipeline.util import _test_classifier, \
    _test_classifier_predict_proba, _test_classifier_iterative_fit
from autosklearn.pipeline.constants import SPARSE

import sklearn.metrics
import numpy as np

from test.test_pipeline.ignored_warnings import ignore_warnings, classifier_warnings


class BaseClassificationComponentTest(unittest.TestCase):
    # Magic command to not run tests on base class
    __test__ = False

    res = None

    module = None
    sk_module = None
    # Hyperparameter which is increased by iterative_fit
    step_hyperparameter = None

    def test_default_iris(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        for i in range(2):
            predictions, targets, n_calls = \
                _test_classifier(dataset="iris",
                                 classifier=self.module)
            self.assertAlmostEqual(self.res["default_iris"],
                                   sklearn.metrics.accuracy_score(targets,
                                                                  predictions),
                                   places=self.res.get(
                                           "default_iris_places", 7))

            if self.res.get("iris_n_calls"):
                self.assertEqual(self.res["iris_n_calls"], n_calls)

    def test_get_max_iter(self):
        if self.__class__ == BaseClassificationComponentTest:
            return

        if not hasattr(self.module, 'iterative_fit'):
            return

        self.module.get_max_iter()

    def test_default_iris_iterative_fit(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if not hasattr(self.module, 'iterative_fit'):
            return

        for i in range(2):
            predictions, targets, classifier = \
                _test_classifier_iterative_fit(dataset="iris",
                                               classifier=self.module)
            self.assertAlmostEqual(self.res["default_iris_iterative"],
                                   sklearn.metrics.accuracy_score(targets,
                                                                  predictions),
                                   places=self.res.get(
                                           "default_iris_iterative_places", 7))

            if self.step_hyperparameter is not None:
                self.assertEqual(
                    getattr(classifier.estimator, self.step_hyperparameter['name']),
                    self.res.get("iris_iterative_n_iter", self.step_hyperparameter['value'])
                )

    def test_default_iris_predict_proba(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        for _ in range(2):
            predictions, targets = _test_classifier_predict_proba(
                dataset="iris", classifier=self.module
            )
            self.assertAlmostEqual(
                self.res["default_iris_proba"],
                sklearn.metrics.log_loss(targets, predictions),
                places=self.res.get("default_iris_proba_places", 7)
            )

    def test_default_iris_sparse(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if SPARSE not in self.module.get_properties()["input"]:
            return

        for i in range(2):
            predictions, targets, _ = \
                _test_classifier(dataset="iris",
                                 classifier=self.module,
                                 sparse=True)
            self.assertAlmostEqual(self.res["default_iris_sparse"],
                                   sklearn.metrics.accuracy_score(targets,
                                                                  predictions),
                                   places=self.res.get(
                                           "default_iris_sparse_places", 7))

    def test_default_digits_binary(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        for i in range(2):
            predictions, targets, _ = \
                _test_classifier(classifier=self.module,
                                 dataset='digits', sparse=False,
                                 make_binary=True)
            self.assertAlmostEqual(self.res["default_digits_binary"],
                                   sklearn.metrics.accuracy_score(
                                       targets, predictions),
                                   places=self.res.get(
                                           "default_digits_binary_places", 7))

    def test_default_digits(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        for i in range(2):
            predictions, targets, n_calls = \
                _test_classifier(dataset="digits",
                                 classifier=self.module)
            self.assertAlmostEqual(self.res["default_digits"],
                                   sklearn.metrics.accuracy_score(targets,
                                                                  predictions),
                                   places=self.res.get(
                                           "default_digits_places", 7))

            if self.res.get("digits_n_calls"):
                self.assertEqual(self.res["digits_n_calls"], n_calls)

    def test_default_digits_iterative_fit(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if not hasattr(self.module, 'iterative_fit'):
            return

        for i in range(2):
            predictions, targets, classifier = \
                _test_classifier_iterative_fit(dataset="digits",
                                               classifier=self.module)
            self.assertAlmostEqual(self.res["default_digits_iterative"],
                                   sklearn.metrics.accuracy_score(targets,
                                                                  predictions),
                                   places=self.res.get(
                                           "default_digits_iterative_places", 7))

            if self.step_hyperparameter is not None:
                self.assertEqual(
                    getattr(classifier.estimator, self.step_hyperparameter['name']),
                    self.res.get("digits_iterative_n_iter", self.step_hyperparameter['value'])
                )

    def test_default_digits_multilabel(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if not self.module.get_properties()["handles_multilabel"]:
            return

        for _ in range(2):
            predictions, targets, _ = _test_classifier(
                classifier=self.module, dataset='digits', make_multilabel=True
            )

            score = sklearn.metrics.precision_score(
                targets, predictions, average='macro', zero_division=0
            )
            self.assertAlmostEqual(
                self.res["default_digits_multilabel"], score,
                places=self.res.get("default_digits_multilabel_places", 7)
            )

    def test_default_digits_multilabel_predict_proba(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if not self.module.get_properties()["handles_multilabel"]:
            return

        for i in range(2):
            predictions, targets = \
                _test_classifier_predict_proba(classifier=self.module,
                                               make_multilabel=True)
            self.assertEqual(predictions.shape, ((50, 3)))
            self.assertAlmostEqual(self.res["default_digits_multilabel_proba"],
                                   sklearn.metrics.roc_auc_score(
                                       targets, predictions, average='macro'),
                                   places=self.res.get(
                                           "default_digits_multilabel_proba_places", 7))

    def test_target_algorithm_multioutput_multiclass_support(self):

        if self.__class__ == BaseClassificationComponentTest:
            return

        if not self.module.get_properties()["handles_multiclass"]:
            return
        elif self.sk_module is not None:
            cls = self.sk_module
            X = np.random.random((10, 10))
            y = np.random.randint(0, 1, size=(10, 10))
            self.assertRaisesRegex(
                ValueError,
                'bad input shape \\(10, 10\\)',
                cls.fit,
                X,
                y
            )
        else:
            return

    def test_module_idempotent(self):
        """ Fitting twice with the same config gives the same model params.

            This is only valid when the random_state passed is an int. If a
            RandomState object is passed then repeated calls to fit will have
            different results. See the section on "Controlling Randomness" in the
            sklearn docs.

            https://scikit-learn.org/0.24/common_pitfalls.html#controlling-randomness
        """
        if self.__class__ == BaseClassificationComponentTest:
            return

        classifier_cls = self.module

        X = np.array([
            [0, 0], [0, 1], [1, 0], [1, 1],
            [0, 0], [0, 1], [1, 0], [1, 1],
            [0, 0], [0, 1], [1, 0], [1, 1],
            [0, 0], [0, 1], [1, 0], [1, 1],
        ])
        y = np.array([
            0, 1, 1, 0,
            0, 1, 1, 0,
            0, 1, 1, 0,
            0, 1, 1, 0,
        ])

        # There are certain errors we ignore so we wrap this in a function
        def fitted_params(model) -> Optional[Dict]:
            """
            Returns the params if fitted successfully, else None if an
            acceptable error occurs
            """
            # We are okay with Numerical in Quadractic disciminant analysis
            def is_QDA_error(err):
                return "Numerical problems in QDA" in err.args[0]

            # We are okay if the BaseClassifier in AdaBoostClassifier is worse
            # than random so no ensemble can be fit
            def is_AdaBoostClassifier_error(err):
                return ("BaseClassifier in AdaBoostClassifier ensemble is worse"
                        + " than random, ensemble can not be fit." in err.args[0])

            def is_unset_param_raw_predictions_val_error(err):
                return ("local variable 'raw_predictions_val' referenced before"
                        + " assignment" in err.args[0])

            try:
                with ignore_warnings(classifier_warnings):
                    model.fit(X.copy(), y.copy())
            except ValueError as e:
                if is_AdaBoostClassifier_error(e) or is_QDA_error(e):
                    return None
            except UnboundLocalError as e:
                if is_unset_param_raw_predictions_val_error(e):
                    return None

            return model.estimator.get_params()

        # We ignore certain keys when comparing
        param_keys_ignored = ['base_estimator']

        # We use the default config + sampled ones
        configuration_space = classifier_cls.get_hyperparameter_search_space()

        default = configuration_space.get_default_configuration()
        sampled = [configuration_space.sample_configuration() for _ in range(2)]

        for seed, config in enumerate([default] + sampled):
            model_args = {"random_state": seed, **config}
            classifier = classifier_cls(**model_args)

            # Get the parameters on the first and second fit with config params
            params_first = fitted_params(classifier)
            if hasattr(classifier.estimator, 'random_state'):
                rs_1 = classifier.random_state
                rs_estimator_1 = classifier.estimator.random_state

            params_second = fitted_params(classifier)
            if hasattr(classifier.estimator, 'random_state'):
                rs_2 = classifier.random_state
                rs_estimator_2 = classifier.estimator.random_state

            # An acceptable error occured, skip to next sample
            if params_first is None or params_second is None:
                continue

            # Remove keys we don't wish to include in the comparison
            for params in [params_first, params_second]:
                for key in param_keys_ignored:
                    if key in params:
                        del params[key]

            # They should have equal parameters
            self.assertEqual(params_first, params_second,
                             f"Failed with model args {model_args}")
            if hasattr(classifier.estimator, 'random_state'):
                assert all([
                    seed == random_state
                    for random_state in [rs_1, rs_estimator_1, rs_2, rs_estimator_2]
                ])



from collections import OrderedDict
import os

from typing import Dict, Optional

from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter

from sklearn.base import BaseEstimator

from ...base import AutoSklearnPreprocessingAlgorithm, find_components, \
    ThirdPartyComponents, AutoSklearnChoice, _addons
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.data_preprocessing.rescaling.abstract_rescaling import (
    Rescaling
)

rescaling_directory = os.path.split(__file__)[0]
_rescalers = find_components(__package__,
                             rescaling_directory,
                             AutoSklearnPreprocessingAlgorithm)
additional_components = ThirdPartyComponents(AutoSklearnPreprocessingAlgorithm)
_addons['data_preprocessing.rescaling'] = additional_components


def add_rescaler(rescaler: Rescaling) -> None:
    additional_components.add_component(rescaler)


class RescalingChoice(AutoSklearnChoice):

    @classmethod
    def get_components(cls: BaseEstimator) -> Dict[str, BaseEstimator]:
        components: Dict[str, BaseEstimator] = OrderedDict()
        components.update(_rescalers)
        components.update(additional_components.components)
        return components

    def get_hyperparameter_search_space(
        self,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
        default: Optional[str] = None,
        include: Optional[Dict[str, str]] = None,
        exclude: Optional[Dict[str, str]] = None,
    ) -> ConfigurationSpace:
        cs = ConfigurationSpace()

        if dataset_properties is None:
            dataset_properties = {}

        # Compile a list of legal preprocessors for this problem
        available_preprocessors = self.get_available_components(
            dataset_properties=dataset_properties,
            include=include, exclude=exclude)

        if len(available_preprocessors) == 0:
            raise ValueError(
                "No rescalers found, please add any rescaling component.")

        if default is None:
            defaults = ['standardize', 'none', 'minmax', 'normalize']
            for default_ in defaults:
                if default_ in available_preprocessors:
                    default = default_
                    break

        preprocessor = CategoricalHyperparameter('__choice__',
                                                 list(
                                                     available_preprocessors.keys()),
                                                 default_value=default)
        cs.add_hyperparameter(preprocessor)
        for name in available_preprocessors:
            preprocessor_configuration_space = available_preprocessors[name]. \
                get_hyperparameter_search_space(dataset_properties)
            parent_hyperparameter = {'parent': preprocessor, 'value': name}
            cs.add_configuration_space(name, preprocessor_configuration_space,
                                       parent_hyperparameter=parent_hyperparameter)

        self.configuration_space = cs
        self.dataset_properties = dataset_properties
        return cs

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        return self.choice.transform(X)



import numpy as np

import pandas as pd
from pandas.api.types import is_numeric_dtype

import pytest

from scipy import sparse

import sklearn.datasets
import sklearn.model_selection

from autosklearn.data.feature_validator import FeatureValidator


# Fixtures to be used in this class. By default all elements have 100 datapoints
@pytest.fixture
def input_data_featuretest(request):
    if request.param == 'numpy_categoricalonly_nonan':
        return np.random.randint(10, size=(100, 10))
    elif request.param == 'numpy_numericalonly_nonan':
        return np.random.uniform(10, size=(100, 10))
    elif request.param == 'numpy_mixed_nonan':
        return np.column_stack([
            np.random.uniform(10, size=(100, 3)),
            np.random.randint(10, size=(100, 3)),
            np.random.uniform(10, size=(100, 3)),
            np.random.randint(10, size=(100, 1)),
        ])
    elif request.param == 'numpy_string_nonan':
        return np.array([
            ['a', 'b', 'c', 'a', 'b', 'c'],
            ['a', 'b', 'd', 'r', 'b', 'c'],
        ])
    elif request.param == 'numpy_categoricalonly_nan':
        array = np.random.randint(10, size=(100, 10)).astype('float')
        array[50, 0:5] = np.nan
        return array
    elif request.param == 'numpy_numericalonly_nan':
        array = np.random.uniform(10, size=(100, 10)).astype('float')
        array[50, 0:5] = np.nan
        # Somehow array is changed to dtype object after np.nan
        return array.astype('float')
    elif request.param == 'numpy_mixed_nan':
        array = np.column_stack([
            np.random.uniform(10, size=(100, 3)),
            np.random.randint(10, size=(100, 3)),
            np.random.uniform(10, size=(100, 3)),
            np.random.randint(10, size=(100, 1)),
        ])
        array[50, 0:5] = np.nan
        return array
    elif request.param == 'numpy_string_nan':
        return np.array([
            ['a', 'b', 'c', 'a', 'b', 'c'],
            [np.nan, 'b', 'd', 'r', 'b', 'c'],
        ])
    elif request.param == 'pandas_categoricalonly_nonan':
        return pd.DataFrame([
            {'A': 1, 'B': 2},
            {'A': 3, 'B': 4},
        ], dtype='category')
    elif request.param == 'pandas_numericalonly_nonan':
        return pd.DataFrame([
            {'A': 1, 'B': 2},
            {'A': 3, 'B': 4},
        ], dtype='float')
    elif request.param == 'pandas_mixed_nonan':
        frame = pd.DataFrame([
            {'A': 1, 'B': 2},
            {'A': 3, 'B': 4},
        ], dtype='category')
        frame['B'] = pd.to_numeric(frame['B'])
        return frame
    elif request.param == 'pandas_categoricalonly_nan':
        return pd.DataFrame([
            {'A': 1, 'B': 2, 'C': np.nan},
            {'A': 3, 'C': np.nan},
        ], dtype='category')
    elif request.param == 'pandas_numericalonly_nan':
        return pd.DataFrame([
            {'A': 1, 'B': 2, 'C': np.nan},
            {'A': 3, 'C': np.nan},
        ], dtype='float')
    elif request.param == 'pandas_mixed_nan':
        frame = pd.DataFrame([
            {'A': 1, 'B': 2, 'C': 8},
            {'A': 3, 'B': 4},
        ], dtype='category')
        frame['B'] = pd.to_numeric(frame['B'])
        return frame
    elif request.param == 'pandas_string_nonan':
        return pd.DataFrame([
            {'A': 1, 'B': 2},
            {'A': 3, 'B': 4},
        ], dtype='string')
    elif request.param == 'list_categoricalonly_nonan':
        return [
            ['a', 'b', 'c', 'd'],
            ['e', 'f', 'c', 'd'],
        ]
    elif request.param == 'list_numericalonly_nonan':
        return [
            [1, 2, 3, 4],
            [5, 6, 7, 8]
        ]
    elif request.param == 'list_mixed_nonan':
        return [
            ['a', 2, 3, 4],
            ['b', 6, 7, 8]
        ]
    elif request.param == 'list_categoricalonly_nan':
        return [
            ['a', 'b', 'c', np.nan],
            ['e', 'f', 'c', 'd'],
        ]
    elif request.param == 'list_numericalonly_nan':
        return [
            [1, 2, 3, np.nan],
            [5, 6, 7, 8]
        ]
    elif request.param == 'list_mixed_nan':
        return [
            ['a', np.nan, 3, 4],
            ['b', 6, 7, 8]
        ]
    elif 'sparse' in request.param:
        # We expect the names to be of the type sparse_csc_nonan
        sparse_, type_, nan_ = request.param.split('_')
        if 'nonan' in nan_:
            data = np.ones(3)
        else:
            data = np.array([1, 2, np.nan])

        # Then the type of sparse
        row_ind = np.array([0, 1, 2])
        col_ind = np.array([1, 2, 1])
        if 'csc' in type_:
            return sparse.csc_matrix((data, (row_ind, col_ind)))
        elif 'csr' in type_:
            return sparse.csr_matrix((data, (row_ind, col_ind)))
        elif 'coo' in type_:
            return sparse.coo_matrix((data, (row_ind, col_ind)))
        elif 'bsr' in type_:
            return sparse.bsr_matrix((data, (row_ind, col_ind)))
        elif 'lil' in type_:
            return sparse.lil_matrix((data))
        elif 'dok' in type_:
            return sparse.dok_matrix(np.vstack((data, data, data)))
        elif 'dia' in type_:
            return sparse.dia_matrix(np.vstack((data, data, data)))
        else:
            ValueError("Unsupported indirect fixture {}".format(request.param))
    elif 'openml' in request.param:
        _, openml_id = request.param.split('_')
        X, y = sklearn.datasets.fetch_openml(data_id=int(openml_id),
                                             return_X_y=True, as_frame=True)
        return X
    else:
        ValueError("Unsupported indirect fixture {}".format(request.param))


# Actual checks for the features
@pytest.mark.parametrize(
    'input_data_featuretest',
    (
        'numpy_categoricalonly_nonan',
        'numpy_numericalonly_nonan',
        'numpy_mixed_nonan',
        'numpy_categoricalonly_nan',
        'numpy_numericalonly_nan',
        'numpy_mixed_nan',
        'pandas_categoricalonly_nonan',
        'pandas_numericalonly_nonan',
        'pandas_mixed_nonan',
        'pandas_numericalonly_nan',
        'list_numericalonly_nonan',
        'list_numericalonly_nan',
        'sparse_bsr_nonan',
        'sparse_bsr_nan',
        'sparse_coo_nonan',
        'sparse_coo_nan',
        'sparse_csc_nonan',
        'sparse_csc_nan',
        'sparse_csr_nonan',
        'sparse_csr_nan',
        'sparse_dia_nonan',
        'sparse_dia_nan',
        'sparse_dok_nonan',
        'sparse_dok_nan',
        'sparse_lil_nonan',
        'sparse_lil_nan',
        'openml_40981',  # Australian
    ),
    indirect=True
)
def test_featurevalidator_supported_types(input_data_featuretest):
    validator = FeatureValidator()
    validator.fit(input_data_featuretest, input_data_featuretest)
    transformed_X = validator.transform(input_data_featuretest)
    if sparse.issparse(input_data_featuretest):
        assert sparse.issparse(transformed_X)
    elif isinstance(input_data_featuretest, list):
        assert isinstance(transformed_X, pd.DataFrame)
    else:
        assert isinstance(transformed_X, type(input_data_featuretest))
    assert np.shape(input_data_featuretest) == np.shape(transformed_X)
    assert validator._is_fitted


@pytest.mark.parametrize(
    'input_data_featuretest',
    (
        'numpy_string_nonan',
        'numpy_string_nan',
    ),
    indirect=True
)
def test_featurevalidator_unsupported_numpy(input_data_featuretest):
    validator = FeatureValidator()
    with pytest.raises(ValueError, match=r".*When providing a numpy array.*not supported."):
        validator.fit(input_data_featuretest)


@pytest.mark.parametrize(
    'input_data_featuretest',
    (
        'numpy_categoricalonly_nonan',
        'numpy_mixed_nonan',
        'numpy_categoricalonly_nan',
        'numpy_mixed_nan',
        'pandas_categoricalonly_nonan',
        'pandas_mixed_nonan',
        'sparse_bsr_nonan',
        'sparse_bsr_nan',
        'sparse_coo_nonan',
        'sparse_coo_nan',
        'sparse_csc_nonan',
        'sparse_csc_nan',
        'sparse_csr_nonan',
        'sparse_csr_nan',
        'sparse_dia_nonan',
        'sparse_dia_nan',
        'sparse_dok_nonan',
        'sparse_dok_nan',
        'sparse_lil_nonan',
    ),
    indirect=True
)
def test_featurevalidator_fitontypeA_transformtypeB(input_data_featuretest):
    """
    Check if we can fit in a given type (numpy) yet transform
    if the user changes the type (pandas then)
    """
    validator = FeatureValidator()
    validator.fit(input_data_featuretest, input_data_featuretest)
    if isinstance(input_data_featuretest, pd.DataFrame):
        complementary_type = input_data_featuretest.to_numpy()
    elif isinstance(input_data_featuretest, np.ndarray):
        complementary_type = pd.DataFrame(input_data_featuretest)
    elif isinstance(input_data_featuretest, list):
        complementary_type = pd.DataFrame(input_data_featuretest)
    elif sparse.issparse(input_data_featuretest):
        complementary_type = sparse.csr_matrix(input_data_featuretest.todense())
    else:
        raise ValueError(type(input_data_featuretest))
    transformed_X = validator.transform(complementary_type)
    assert np.shape(input_data_featuretest) == np.shape(transformed_X)
    assert validator._is_fitted


def test_featurevalidatorget_feat_type_from_columns():
    """
    Makes sure that encoded columns are returned by get_feat_type_from_columns
    whereas numerical columns are not returned
    """
    validator = FeatureValidator()

    df = pd.DataFrame([
        {'int': 1, 'float': 1.0, 'category': 'one', 'bool': True},
        {'int': 2, 'float': 2.0, 'category': 'two', 'bool': False},
    ])

    for col in df.columns:
        df[col] = df[col].astype(col)

    feature_types = validator.get_feat_type_from_columns(df)

    assert feature_types == {'int': 'numerical',
                             'float': 'numerical',
                             'category': 'categorical',
                             'bool': 'categorical'}


def test_features_unsupported_calls_are_raised():
    """
    Makes sure we raise a proper message to the user,
    when providing not supported data input or using the validator in a way that is not
    expected
    """
    validator = FeatureValidator()
    with pytest.raises(ValueError, match=r"Auto-sklearn does not support time"):
        validator.fit(
            pd.DataFrame({'datetime': [pd.Timestamp('20180310')]})
        )
    with pytest.raises(ValueError, match=r"Auto-sklearn only supports.*yet, the provided input"):
        validator.fit({'input1': 1, 'input2': 2})
    validator = FeatureValidator()
    with pytest.raises(ValueError, match=r"has unsupported dtype string"):
        validator.fit(pd.DataFrame([{'A': 1, 'B': 2}], dtype='string'))
    with pytest.raises(ValueError, match=r"The feature dimensionality of the train and test"):
        validator.fit(X_train=np.array([[1, 2, 3], [4, 5, 6]]),
                      X_test=np.array([[1, 2, 3, 4], [4, 5, 6, 7]]),
                      )
    with pytest.raises(ValueError, match=r"Cannot call transform on a validator that is not fit"):
        validator.transform(np.array([[1, 2, 3], [4, 5, 6]]))
    validator = FeatureValidator(feat_type=['Numerical'])
    with pytest.raises(ValueError, match=r"providing the option feat_type to the fit method is.*"):
        validator.fit(pd.DataFrame([[1, 2, 3], [4, 5, 6]]))
    with pytest.raises(ValueError, match=r"feat_type does not have same number of.*"):
        validator.fit(np.array([[1, 2, 3], [4, 5, 6]]))
    validator = FeatureValidator(feat_type=[1, 2, 3])
    with pytest.raises(ValueError, match=r"feat_type must only contain strings.*"):
        validator.fit(np.array([[1, 2, 3], [4, 5, 6]]))
    validator = FeatureValidator(feat_type=['1', '2', '3'])
    with pytest.raises(ValueError, match=r"Only `Categorical` and `Numerical` are.*"):
        validator.fit(np.array([[1, 2, 3], [4, 5, 6]]))


def test_no_new_category_after_fit():
    """
    This test makes sure that we can actually pass new categories to the estimator
    without throwing an error
    """
    # Then make sure we catch categorical extra categories
    x = pd.DataFrame({'A': [1, 2, 3, 4], 'B': [5, 6, 7, 8]}, dtype='category')
    validator = FeatureValidator()
    validator.fit(x)
    x['A'] = x['A'].apply(lambda x: x*x)
    validator.transform(x)


# Actual checks for the features
@pytest.mark.parametrize(
    'openml_id',
    (
        40981,  # Australian
        3,  # kr-vs-kp
        1468,  # cnae-9
        40975,  # car
        40984,  # Segment
    ),
)
@pytest.mark.parametrize('train_data_type', ('numpy', 'pandas', 'list'))
@pytest.mark.parametrize('test_data_type', ('numpy', 'pandas', 'list'))
def test_featurevalidator_new_data_after_fit(openml_id,
                                             train_data_type, test_data_type):

    # List is currently not supported as infer_objects
    # cast list objects to type objects
    if train_data_type == 'list' or test_data_type == 'list':
        pytest.skip()

    validator = FeatureValidator()

    if train_data_type == 'numpy':
        X, y = sklearn.datasets.fetch_openml(data_id=openml_id,
                                             return_X_y=True, as_frame=False)
    elif train_data_type == 'pandas':
        X, y = sklearn.datasets.fetch_openml(data_id=openml_id,
                                             return_X_y=True, as_frame=True)
    else:
        X, y = sklearn.datasets.fetch_openml(data_id=openml_id,
                                             return_X_y=True, as_frame=True)
        X = X.values.tolist()
        y = y.values.tolist()

    X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(
        X, y, random_state=1)

    validator.fit(X_train)

    transformed_X = validator.transform(X_test)

    # Basic Checking
    if sparse.issparse(input_data_featuretest):
        assert sparse.issparse(transformed_X)
    elif isinstance(input_data_featuretest, list):
        assert isinstance(transformed_X, pd.DataFrame)
    else:
        assert isinstance(transformed_X, type(X_train))
    assert np.shape(X_test) == np.shape(transformed_X)


@pytest.mark.parametrize(
    'openml_id',
    (
        40981,  # Australian
        3,  # kr-vs-kp
        1468,  # cnae-9
        40975,  # car
        40984,  # Segment
        2,  # anneal
    ),
)
def test_list_to_dataframe(openml_id):

    X_pandas, y_pandas = sklearn.datasets.fetch_openml(data_id=openml_id,
                                                       return_X_y=True, as_frame=True)
    X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(
        X_pandas, y_pandas, random_state=1)

    X_list = X_train.values.tolist()
    validator = FeatureValidator()
    validator.fit(X_list)
    transformed_X = validator.transform(X_list)
    for i, col in enumerate(X_pandas.columns):
        if is_numeric_dtype(X_pandas[col].dtype):
            # convert dtype translates 72.0 to 72. Be robust against this!
            assert is_numeric_dtype(transformed_X[i].dtype)
        else:
            assert X_pandas[col].dtype.name == transformed_X[i].dtype.name, col

    # Also make sure that at testing time
    # this work
    transformed_X = validator.transform(X_test.values.tolist())
    for i, col in enumerate(X_pandas.columns):
        if is_numeric_dtype(X_pandas[col].dtype):
            # convert dtype translates 72.0 to 72. Be robust against this!
            assert is_numeric_dtype(transformed_X[i].dtype)
        else:
            assert X_pandas[col].dtype.name == transformed_X[i].dtype.name, col


@pytest.mark.parametrize(
    'input_data_featuretest',
    (
        'sparse_bsr_nonan',
        'sparse_bsr_nan',
        'sparse_coo_nonan',
        'sparse_coo_nan',
        'sparse_csc_nonan',
        'sparse_csc_nan',
        'sparse_csr_nonan',
        'sparse_csr_nan',
        'sparse_dia_nonan',
        'sparse_dia_nan',
        'sparse_dok_nonan',
        'sparse_dok_nan',
        'sparse_lil_nonan',
        'sparse_lil_nan',
    ),
    indirect=True
)
def test_sparse_output_is_csr(input_data_featuretest):
    validator = FeatureValidator()
    validator.fit(input_data_featuretest, input_data_featuretest)
    transformed_X = validator.transform(input_data_featuretest)
    assert sparse.issparse(transformed_X)
    assert isinstance(transformed_X, sparse.csr_matrix)


def test_unsupported_dataframe_sparse():
    df = pd.DataFrame({'A': pd.Series(pd.arrays.SparseArray(np.random.randn(10)))})
    validator = FeatureValidator()
    with pytest.raises(ValueError, match=r"Auto-sklearn does not yet support sparse pandas"):
        validator.fit(df)
